# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2025/4/8 14:16
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: claude_official_llm_channel.py
from typing import Optional

from agentuniverse.llm.llm_channel.llm_channel import LLMChannel

CLAUDE_MAX_CONTEXT_LENGTH = {
    "claude-3-opus-20240229": 200000,
    "claude-3-sonnet-20240229": 200000,
    "claude-3-haiku-20240307": 200000,
    "claude-2.1": 200000,
    "claude-2.0": 100000,
    "claude-instant-1.2": 100000
}


class ClaudeOfficialLLMChannel(LLMChannel):
    channel_api_base: Optional[str] = "https://api.anthropic.com/v1/"

    def max_context_length(self) -> int:
        if super().max_context_length():
            return super().max_context_length()
        return CLAUDE_MAX_CONTEXT_LENGTH.get(self.channel_model_name, 8000)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-
import copy
from typing import List, Optional, Any

from agentuniverse.agent.action.tool.tool_manager import ToolManager
from agentuniverse.base.component.component_base import ComponentBase
from agentuniverse.base.component.component_enum import ComponentEnum
from agentuniverse.base.config.application_configer.application_config_manager import \
    ApplicationConfigManager
from agentuniverse.base.config.component_configer.component_configer import \
    ComponentConfiger


class Toolkit(ComponentBase):
    """
    The toolkit is a collection of tools.

    Attributes:
        name (str): The name of the toolkit.
        description (str): The description of the tool.
        include (list):
    """

    name: str = ""
    description: Optional[str] = None
    include: Optional[List[str]] = []
    as_mcp_tool: Any = None

    def __init__(self, **kwargs):
        super().__init__(component_type=ComponentEnum.TOOLKIT, **kwargs)

    @property
    def tool_names(self) -> list:
        """Return all tool names in toolkit."""
        return copy.deepcopy(self.include)


    @property
    def tool_descriptions(self) -> list:
        """Return all tools' descriptions in toolkit."""
        tools = [ToolManager().get_instance_obj(tool_name, new_instance=False) for tool_name in self.include]
        tools_descriptions = [f'tool name:{tool.name}\ntool description:{tool.description}\n' for tool in tools]
        return tools_descriptions

    @property
    def func_call_list(self) -> list:
        raise NotImplementedError

    def get_instance_code(self) -> str:
        """Return the full name of the toolkit."""
        appname = ApplicationConfigManager().app_configer.base_info_appname
        return f'{appname}.{self.component_type.value.lower()}.{self.name}'

    def initialize_by_component_configer(self, component_configer: ComponentConfiger) -> 'Toolkit':
        """Initialize the Toolkit by the ComponentConfiger object.
        Args:
            component_configer(LLMConfiger): the ComponentConfiger object
        Returns:
            Toolkit: the Toolkit object
        """
        try:
            for key, value in component_configer.configer.value.items():
                if key != 'metadata':
                    setattr(self, key, value)
        except Exception as e:
            print(f"Error during configuration initialization: {str(e)}")
        if component_configer.name:
            self.name = component_configer.name
        if component_configer.description:
            self.description = component_configer.description
        if hasattr(component_configer, "include"):
            self.include = component_configer.include
        self._initialize_by_component_configer(component_configer)
        return self



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/26 14:31
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: thread_with_result.py

import weakref
from concurrent.futures.thread import ThreadPoolExecutor, _worker, \
    _threads_queues
from threading import Thread

from agentuniverse.base.context.context_coordinator import ContextCoordinator


class ThreadWithReturnValue(Thread):
    """A thread can save the target func exec result."""

    def __init__(self, group=None, target=None, name=None,
                 args=(), kwargs=None):
        super().__init__(group, target, name, args, kwargs)

        if kwargs is None:
            kwargs = {}
        self.kwargs = kwargs
        self.args = args
        self.target = target
        self._return = None
        self.error = None
        self._context_pack = ContextCoordinator.save_context()

    def run(self):
        """Run the target func and save result in _return."""
        if self.target is not None:
            # set the context values in the thread
            ContextCoordinator.recover_context(self._context_pack)
            try:
                self._return = self.target(*self.args, **self.kwargs)
            except Exception as e:
                self.error = e

    def result(self):
        """Wait for target func finished, then return the result or raise an
        error."""
        self.join()
        if self.error is not None:
            raise self.error
        return self._return


class ThreadPoolExecutorWithReturnValue(ThreadPoolExecutor):

    def _adjust_thread_count(self):
        # if idle threads are available, don't spin new threads
        if self._idle_semaphore.acquire(timeout=0):
            return

        # When the executor gets lost, the weakref callback will wake up
        # the worker threads.
        def weakref_cb(_, q=self._work_queue):
            q.put(None)

        num_threads = len(self._threads)
        if num_threads < self._max_workers:
            thread_name = '%s_%d' % (self._thread_name_prefix or self,
                                     num_threads)
            t = ThreadWithReturnValue(name=thread_name, target=_worker,
                                 args=(weakref.ref(self, weakref_cb),
                                       self._work_queue,
                                       self._initializer,
                                       self._initargs))
            t.start()
            self._threads.add(t)
            _threads_queues[t] = self._work_queue


# !/usr/bin/env python3
# -*- coding:utf-8 -*-
# @Time    : 2024/3/12 16:17
# @Author  : jerry.zzw 
# @Email   : jerry.zzw@antgroup.com
# @FileName: app_configer.py
import importlib
from typing import Optional, Dict, Any
from agentuniverse.base.config.component_configer.component_configer import \
    ComponentConfiger
from agentuniverse.base.config.component_configer.configers.llm_configer import LLMConfiger
from agentuniverse.base.config.component_configer.configers.tool_configer import ToolConfiger
from agentuniverse.base.config.configer import Configer
from agentuniverse.base.config.custom_configer.default_llm_configer import DefaultLLMConfiger


class AppConfiger(object):
    """The AppConfiger class, which is used to load and manage the application configuration."""

    def __init__(self):
        """Initialize the AppConfiger."""
        self.__configer: Optional[Configer] = None
        self.__base_info_appname: Optional[str] = None
        self.__core_default_package_list: Optional[list[str]] = None
        self.__core_agent_package_list: Optional[list[str]] = None
        self.__core_knowledge_package_list: Optional[list[str]] = None
        self.__core_llm_package_list: Optional[list[str]] = None
        self.__core_planner_package_list: Optional[list[str]] = None
        self.__core_tool_package_list: Optional[list[str]] = None
        self.__core_toolkit_package_list: Optional[list[str]] = None
        self.__core_memory_package_list: Optional[list[str]] = None
        self.__core_service_package_list: Optional[list[str]] = None
        self.__core_sqldb_wrapper_package_list: Optional[list[str]] = None
        self.__core_prompt_package_list: Optional[list[str]] = None
        self.__core_product_package_list: Optional[list[str]] = None
        self.__core_workflow_package_list: Optional[list[str]] = None
        self.__core_embedding_package_list: Optional[list[str]] = None
        self.__core_doc_processor_package_list: Optional[list[str]] = None
        self.__core_reader_package_list: Optional[list[str]] = None
        self.__core_store_package_list: Optional[list[str]] = None
        self.__core_rag_router_package_list: Optional[list[str]] = None
        self.__core_query_paraphraser_package_list: Optional[list[str]] = None
        self.__core_memory_compressor_package_list: Optional[list[str]] = None
        self.__core_memory_storage_package_list: Optional[list[str]] = None
        self.__core_work_pattern_package_list: Optional[list[str]] = None
        self.__core_log_sink_package_list: Optional[list[str]] = None
        self.__core_llm_channel_package_list: Optional[list[str]] = None
        self.__conversation_memory_configer: Optional[dict] = {}
        self.__root_package_name: Optional[str] = None
        self.__yaml_func_instance = None
        self.__default_llm_configer: DefaultLLMConfiger = None
        self.__tool_configer_map: Dict[str, ToolConfiger] = {}
        self.__toolkit_configer_map: Dict[str, ComponentConfiger] = {}
        self.__llm_configer_map: Dict[str, LLMConfiger] = {}
        self.__agent_llm_set: Optional[set[str]] = set()
        self.__agent_tool_set: Optional[set[str]] = set()
        self.__agent_toolkit_set: Optional[set[str]] = set()
        self.__llm_plugins: Optional[Any] = set()

    @property
    def base_info_appname(self) -> Optional[str]:
        """Return the appname of the application."""
        return self.__base_info_appname

    @property
    def core_default_package_list(self) -> Optional[list[str]]:
        """Return the default package list of the core."""
        return self.__core_default_package_list

    @property
    def core_agent_package_list(self) -> Optional[list[str]]:
        """Return the agent package list of the core."""
        return self.__core_agent_package_list

    @property
    def core_knowledge_package_list(self) -> Optional[list[str]]:
        """Return the knowledge package list of the core."""
        return self.__core_knowledge_package_list

    @property
    def core_llm_package_list(self) -> Optional[list[str]]:
        """Return the llm package list of the core."""
        return self.__core_llm_package_list

    @property
    def core_planner_package_list(self) -> Optional[list[str]]:
        """Return the planner package list of the core."""
        return self.__core_planner_package_list

    @property
    def core_tool_package_list(self) -> Optional[list[str]]:
        """Return the tool package list of the core."""
        return self.__core_tool_package_list

    @property
    def core_toolkit_package_list(self) -> Optional[list[str]]:
        """Return the toolkit package list of the core."""
        return self.__core_toolkit_package_list

    @property
    def core_memory_package_list(self) -> Optional[list[str]]:
        """Return the memory package list of the core."""
        return self.__core_memory_package_list

    @property
    def core_service_package_list(self) -> Optional[list[str]]:
        """Return the service package list of the core."""
        return self.__core_service_package_list

    @property
    def core_sqldb_wrapper_package_list(self) -> Optional[list[str]]:
        """Return the sql db wrapper package list of the core."""
        return self.__core_sqldb_wrapper_package_list

    @property
    def core_prompt_package_list(self) -> Optional[list[str]]:
        return self.__core_prompt_package_list

    @property
    def core_product_package_list(self) -> Optional[list[str]]:
        return self.__core_product_package_list

    @property
    def core_workflow_package_list(self) -> Optional[list[str]]:
        return self.__core_workflow_package_list

    @property
    def core_embedding_package_list(self) -> Optional[list[str]]:
        """Return the embedding package list of the core."""
        return self.__core_embedding_package_list

    @property
    def core_doc_processor_package_list(self) -> Optional[list[str]]:
        """Return the document processor package list of the core."""
        return self.__core_doc_processor_package_list

    @property
    def core_reader_package_list(self) -> Optional[list[str]]:
        """Return the reader package list of the core."""
        return self.__core_reader_package_list

    @property
    def core_store_package_list(self) -> Optional[list[str]]:
        """Return the store package list of the core."""
        return self.__core_store_package_list

    @property
    def core_rag_router_package_list(self) -> Optional[list[str]]:
        """Return the RAG router package list of the core."""
        return self.__core_rag_router_package_list

    @property
    def core_query_paraphraser_package_list(self) -> Optional[list[str]]:
        """Return the query paraphraser package list of the core."""
        return self.__core_query_paraphraser_package_list

    @property
    def core_memory_compressor_package_list(self) -> Optional[list[str]]:
        """Return the memory compressor package list of the core."""
        return self.__core_memory_compressor_package_list

    @property
    def core_memory_storage_package_list(self) -> Optional[list[str]]:
        """Return the memory storage package list of the core."""
        return self.__core_memory_storage_package_list

    @property
    def core_work_pattern_package_list(self) -> Optional[list[str]]:
        """Return the work pattern package list of the core."""
        return self.__core_work_pattern_package_list

    @property
    def core_log_sink_package_list(self) -> Optional[list[str]]:
        """Return the work pattern package list of the core."""
        return self.__core_log_sink_package_list

    @property
    def core_llm_channel_package_list(self) -> Optional[list[str]]:
        """Return the llm channel package list of the core."""
        return self.__core_llm_channel_package_list

    @property
    def conversation_memory_configer(self) -> dict:
        return self.__conversation_memory_configer

    @property
    def root_package_name(self) -> str:
        return self.__root_package_name

    @property
    def yaml_func_instance(self):
        return self.__yaml_func_instance

    @yaml_func_instance.setter
    def yaml_func_instance(self, value):
        self.__yaml_func_instance = value

    @property
    def default_llm_configer(self) -> DefaultLLMConfiger:
        return self.__default_llm_configer

    @default_llm_configer.setter
    def default_llm_configer(self, value: DefaultLLMConfiger):
        self.__default_llm_configer = value

    @property
    def tool_configer_map(self) -> Dict[str, ToolConfiger]:
        return self.__tool_configer_map

    @tool_configer_map.setter
    def tool_configer_map(self, value: Dict[str, ToolConfiger]):
        self.__tool_configer_map = value

    @property
    def toolkit_configer_map(self) -> Dict[str, ComponentConfiger]:
        return self.__toolkit_configer_map

    @toolkit_configer_map.setter
    def toolkit_configer_map(self, value: Dict[str, ComponentConfiger]):
        self.__toolkit_configer_map = value

    @property
    def llm_configer_map(self) -> Dict[str, LLMConfiger]:
        return self.__llm_configer_map

    @llm_configer_map.setter
    def llm_configer_map(self, value: Dict[str, LLMConfiger]):
        self.__llm_configer_map = value

    @property
    def agent_llm_set(self) -> set:
        return self.__agent_llm_set

    @agent_llm_set.setter
    def agent_llm_set(self, value: set):
        self.__agent_llm_set = value

    @property
    def agent_tool_set(self) -> set:
        return self.__agent_tool_set

    @property
    def agent_toolkit_set(self) -> set:
        return self.__agent_toolkit_set

    @agent_tool_set.setter
    def agent_tool_set(self, value: set):
        self.__agent_tool_set = value

    @property
    def llm_plugins(self):
        return self.__llm_plugins

    @classmethod
    def load_llm_plugins(cls, plugin_modules):
        funcs = []
        for item in plugin_modules:
            module_name, func_name = item.rsplit('.', 1)
            module = importlib.import_module(module_name)
            funcs.append(getattr(module, func_name))
        return funcs

    def load_by_configer(self, configer: Configer) -> 'AppConfiger':
        """Load the AppConfiger by the given Configer.

        Args:
            configer(Configer): the Configer object
        Returns:
            AppConfiger: the AppConfiger object
        """
        self.__configer = configer
        self.__base_info_appname = configer.value.get('BASE_INFO', {}).get('appname')
        self.__root_package_name = configer.value.get('PACKAGE_PATH_INFO', {}).get('ROOT_PACKAGE')
        self.__core_default_package_list = configer.value.get('CORE_PACKAGE', {}).get('default')
        self.__core_agent_package_list = configer.value.get('CORE_PACKAGE', {}).get('agent')
        self.__core_knowledge_package_list = configer.value.get('CORE_PACKAGE', {}).get('knowledge')
        self.__core_llm_package_list = configer.value.get('CORE_PACKAGE', {}).get('llm')
        self.__core_planner_package_list = configer.value.get('CORE_PACKAGE', {}).get('planner')
        self.__core_tool_package_list = configer.value.get('CORE_PACKAGE', {}).get('tool')
        self.__core_toolkit_package_list = configer.value.get('CORE_PACKAGE', {}).get('toolkit')
        self.__core_memory_package_list = configer.value.get('CORE_PACKAGE', {}).get('memory')
        self.__core_service_package_list = configer.value.get('CORE_PACKAGE', {}).get('service')
        self.__core_sqldb_wrapper_package_list = configer.value.get('CORE_PACKAGE', {}).get('sqldb_wrapper')
        self.__core_prompt_package_list = configer.value.get('CORE_PACKAGE', {}).get('prompt')
        self.__core_product_package_list = configer.value.get('CORE_PACKAGE', {}).get('product')
        self.__core_workflow_package_list = configer.value.get('CORE_PACKAGE', {}).get('workflow')
        self.__core_embedding_package_list = configer.value.get('CORE_PACKAGE', {}).get('embedding')
        self.__core_doc_processor_package_list = configer.value.get('CORE_PACKAGE', {}).get('doc_processor')
        self.__core_reader_package_list = configer.value.get('CORE_PACKAGE', {}).get('reader')
        self.__core_store_package_list = configer.value.get('CORE_PACKAGE', {}).get('store')
        self.__core_rag_router_package_list = configer.value.get('CORE_PACKAGE', {}).get('rag_router')
        self.__core_query_paraphraser_package_list = configer.value.get('CORE_PACKAGE', {}).get('query_paraphraser')
        self.__core_memory_compressor_package_list = configer.value.get('CORE_PACKAGE', {}).get('memory_compressor')
        self.__core_memory_storage_package_list = configer.value.get('CORE_PACKAGE', {}).get('memory_storage')
        self.__core_work_pattern_package_list = configer.value.get('CORE_PACKAGE', {}).get('work_pattern')
        self.__core_log_sink_package_list = configer.value.get('CORE_PACKAGE', {}).get('log_sink')
        self.__core_llm_channel_package_list = configer.value.get('CORE_PACKAGE', {}).get('llm_channel')
        self.__conversation_memory_configer = configer.value.get('CONVERSATION_MEMORY', {})
        self.__llm_plugins = self.load_llm_plugins(configer.value.get("PLUGINS", {}).get("llm_plugins", []))
        return self



# !/usr/bin/env python3
# -*- coding:utf-8 -*-
# @Time    :
# @Author  :
# @Email   :
# @FileName: google_search_tool.py

from typing import Optional

from agentuniverse.agent.action.tool.tool import Tool, ToolInput
from agentuniverse.base.util.env_util import get_from_env
from langchain_community.utilities.google_serper import GoogleSerperAPIWrapper
from pydantic import Field


class GoogleSearchTool(Tool):
    """The demo google search tool.

    Implement the execute method of demo google search tool, using the `GoogleSerperAPIWrapper` to implement a simple Google search.

    Note:
        You need to sign up for a free account at https://serper.dev and get the serpher api key (2500 free queries).
    """

    serper_api_key: Optional[str] = Field(default_factory=lambda: get_from_env("SERPER_API_KEY"))

    def execute(self, input: str):
        # get top10 results from Google search.
        search = GoogleSerperAPIWrapper(serper_api_key=self.serper_api_key, k=10, gl="us", hl="en", type="search")
        return search.run(query=input)

    async def async_execute(self, input: str):
        # get top10 results from Google search.
        search = GoogleSerperAPIWrapper(serper_api_key=self.serper_api_key, k=10, gl="us", hl="en", type="search")
        return await search.arun(query=input)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2/16/25 9:50 PM
# @Author  : Xinyuan Xu
# @Email   : xuxinyuan2019@gmail.com
# @FileName: azureopenai_embedding.py

from typing import Any, Optional, List
from pydantic import Field

from openai import AzureOpenAI, AsyncAzureOpenAI

from agentuniverse.base.util.env_util import get_from_env
from agentuniverse.agent.action.knowledge.embedding.embedding import Embedding
from agentuniverse.base.config.component_configer.component_configer import ComponentConfiger


class AzureOpenAIEmbedding(Embedding):
    """The Azure OpenAI embedding class."""

    azure_api_key: Optional[str] = Field(
        default_factory=lambda: get_from_env("AZURE_OPENAI_API_KEY"))
    
    resource_name: Optional[str] = Field(
        default_factory=lambda: get_from_env("AZURE_OPENAI_RESOURCE_NAME"))

    azure_api_version: Optional[str] = Field(
        default_factory=lambda: get_from_env("AZURE_API_VERSION"))

    embedding_model_name: Optional[str] = None
    embedding_dims: Optional[int] = None

    client: Any = None
    async_client: Any = None


    def get_embeddings(self, texts: List[str], **kwargs) -> List[List[float]]:
        """
        Retrieve text embeddings for a list of input texts using Azure OpenAI API.
        Args:
            texts (List[str]): A list of input texts to be embedded.
        Returns:
            List[List[float]]: A list of embeddings corresponding to the input texts.
        Raises:
            Exception: If the API call fails or if required configuration is missing.
        """
        self._initialize_clients()
        
        try:
            if self.embedding_dims is not None:
                response = self.client.embeddings.create(
                    input=texts,
                    model=self.embedding_model_name,
                    dimensions=self.embedding_dims
                )
            else:
                response = self.client.embeddings.create(
                    input=texts,
                    model=self.embedding_model_name
                )
            return [item.embedding for item in response.data]

        except Exception as e:
            raise Exception(f"Failed to get embeddings: {e}")


    async def async_get_embeddings(self, texts: List[str], **kwargs) -> List[List[float]]:
        """
        Retrieve text embeddings for a list of input texts using Azure OpenAI API asynchronously.
        Args:
            texts (List[str]): A list of input texts to be embedded.
        Returns:
            List[List[float]]: A list of embeddings corresponding to the input texts.
        Raises:
            Exception: If the API call fails or if required configuration is missing.
        """
        self._initialize_clients()

        try:
            if self.embedding_dims is not None:
                response = await self.async_client.embeddings.create(
                    input=texts,
                    model=self.embedding_model_name,
                    dimensions=self.embedding_dims
                )
            else:
                response = await self.async_client.embeddings.create(
                    input=texts,
                    model=self.embedding_model_name
                )
            return [item.embedding for item in response.data]

        except Exception as e:
            raise Exception(f"Failed to get embeddings: {e}")


    def as_langchain(self) -> Any:
        """
        Convert the AzureOpenAIEmbedding instance to a LangChainAzureOpenAIEmbedding instance.
        """
        self._initialize_clients()

        from langchain_community.embeddings.azure_openai import AzureOpenAIEmbeddings
        return AzureOpenAIEmbeddings(openai_api_key=self.azure_api_key, client=self.client.embeddings, async_client=self.async_client.embeddings, azure_endpoint=f"https://{self.resource_name}.openai.azure.com/")


    def _initialize_by_component_configer(self, embedding_configer: ComponentConfiger) -> 'Embedding':
        """
        Initialize the embedding by the ComponentConfiger object.
        Args:
            embedding_configer(ComponentConfiger): A configer contains embedding configuration.
        Returns:    
            Embedding: A AzureOpenAIEmbedding instance.
        """
        super()._initialize_by_component_configer(embedding_configer)
        if hasattr(embedding_configer, "azure_api_key"):
            self.azure_api_key = embedding_configer.azure_api_key
        if hasattr(embedding_configer, "resource_name"):
            self.resource_name = embedding_configer.resource_name
        return self


    def _initialize_clients(self) -> None:
        if not self.azure_api_key:
            raise Exception("AZURE_OPENAI_API_KEY is missing")
        if not self.resource_name:
            raise Exception("AZURE_OPENAI_RESOURCE_NAME is missing")
        if not self.azure_api_version:
            raise Exception("AZURE_API_VERSION is missing")

        if self.client is None:
            self.client = AzureOpenAI(
                api_key=self.azure_api_key,
                api_version=self.azure_api_version,
                azure_endpoint=f"https://{self.resource_name}.openai.azure.com"
            )
        if self.async_client is None:
            self.async_client = AsyncAzureOpenAI(
                api_key=self.azure_api_key,
                api_version=self.azure_api_version,
                azure_endpoint=f"https://{self.resource_name}.openai.azure.com"
            )


# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2025/4/8 11:44
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: gemini_official_llm_channel.py
from typing import Optional

from agentuniverse.llm.llm_channel.llm_channel import LLMChannel

GEMINI_MAX_CONTEXT_LENGTH = {
    "gemini-2.0-flash": 1048576
}


class GeminiOfficialLLMChannel(LLMChannel):
    channel_api_base: Optional[str] = "https://generativelanguage.googleapis.com/v1beta/openai/"

    def max_context_length(self) -> int:
        if super().max_context_length():
            return super().max_context_length()
        return GEMINI_MAX_CONTEXT_LENGTH.get(self.channel_model_name, 8000)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    :
# @Author  :
# @Email   :
# @FileName: langchain_tool.py
import importlib
import importlib
from typing import Optional, Type
from agentuniverse.agent.action.tool.tool import Tool, ToolInput
from agentuniverse.base.config.component_configer.configers.tool_configer import ToolConfiger
from langchain_core.tools import BaseTool


class LangChainTool(Tool):
    name: Optional[str] = ""
    description: Optional[str] = ""
    tool: Optional[BaseTool] = None

    def execute(self, input: str, callbacks):
        return self.tool.run(input, callbacks=callbacks)

    async def async_execute(self, input: str, callbacks):
        return await self.tool.arun(input, callbacks=callbacks)

    def initialize_by_component_configer(self, component_configer: ToolConfiger) -> 'Tool':
        super().initialize_by_component_configer(component_configer)
        self.tool = self.init_langchain_tool(component_configer)
        if not component_configer.description and self.tool is not None:
            self.description = self.tool.description
        return self

    def init_langchain_tool(self, component_configer):
        langchain_info = component_configer.configer.value.get('langchain')
        module = langchain_info.get("module")
        class_name = langchain_info.get("class_name")
        module = importlib.import_module(module)
        clz = getattr(module, class_name)
        init_params = langchain_info.get("init_params")
        self.get_langchain_tool(init_params, clz)
        return self.tool

    def get_langchain_tool(self, init_params: dict, clz: Type[BaseTool]):
        if init_params:
            self.tool = clz(**init_params)
        else:
            self.tool = clz()



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2025/2/7 10:00
# @Author  : wangyapei
# @FileName: website_bs4_reader.py

from typing import Dict, List, Optional
from urllib.parse import urljoin, urlparse
import random
import time

from bs4 import BeautifulSoup
import httpx

from agentuniverse.agent.action.knowledge.store.document import Document
from agentuniverse.agent.action.knowledge.reader.reader import Reader
from pydantic import Field


class WebsiteBs4Reader(Reader):
    """Website content reader based on BeautifulSoup
    
    Crawls website content and extracts main text, with support for limiting crawl depth and max links.
    
    Note:
        Required packages:
            - beautifulsoup4: `pip install beautifulsoup4`
            - httpx: `pip install httpx`
    
    Attributes:
        max_depth: Maximum crawl depth, defaults to 1
        max_links: Maximum number of links to crawl, defaults to 1
    """
    
    max_depth: int = Field(default=1, description="Maximum crawl depth")
    max_links: int = Field(default=1, description="Maximum number of links to crawl")
    
    def __init__(self, **data):
        super().__init__(**data)
        self._visited = set()  # Track visited URLs
        self._urls_to_crawl = []  # Queue of URLs to crawl

    def _load_data(self, url: str, ext_info: Optional[Dict] = None) -> List[Document]:
        """Crawl website content and return list of documents
        
        Args:
            url: Website URL to crawl
            ext_info: Optional metadata dict. Can contain max_depth and max_links to override defaults
            
        Returns:
            List[Document]: List of documents containing webpage content
            
        Raises:
            ValueError: If URL is invalid or empty
            ConnectionError: If network connection fails
            RuntimeError: If content parsing fails
        """
        # Validate URL
        if not url or not isinstance(url, str):
            return []
            
        if not url.startswith(('http://', 'https://')):
            raise ValueError("Invalid URL format")

        # Update config from ext_info if provided
        if ext_info:
            if 'max_depth' in ext_info:
                self.max_depth = ext_info['max_depth']
            if 'max_links' in ext_info:
                self.max_links = ext_info['max_links']
                
        try:
            # Crawl website and create documents
            crawler_result = self._crawl_website(url)
            documents = []
            
            for crawled_url, content in crawler_result.items():
                metadata = {"source": crawled_url}
                if ext_info:
                    metadata.update(ext_info)
                documents.append(Document(text=content, metadata=metadata))
            return documents
            
        except httpx.RequestError as e:
            raise ConnectionError(f"Network connection failed: {str(e)}")
        except Exception as e:
            raise RuntimeError(f"Content parsing failed: {str(e)}")

    def _get_primary_domain(self, url: str) -> str:
        """Extract primary domain from URL
        
        Args:
            url: Full URL
            
        Returns:
            str: Primary domain
        """
        domain_parts = urlparse(url).netloc.split(".")
        return ".".join(domain_parts[-2:])

    def _extract_main_content(self, soup: BeautifulSoup) -> str:
        """Extract main content from webpage
        
        Args:
            soup: BeautifulSoup object
            
        Returns:
            str: Extracted text content
        """
        # Generic content extraction strategy
        content_selectors = [
            ('article', None),
            ('main', None),
            ('div', 'content'),
            ('div', 'main-content'),
            ('div', 'post-content'),
            ('div', 'article-content'),
            ('div', lambda x: x and 'content' in x)
        ]

        # Try each selector pattern
        for tag, class_name in content_selectors:
            element = soup.find(tag, class_=class_name)
            if element:
                text = element.get_text(strip=True, separator=" ")
                if len(text) > 200:  # Only return if substantial content found
                    return text

        # Fallback: Extract largest text block
        all_elements = soup.find_all(['div', 'section'])
        if all_elements:
            return max(
                [e.get_text(strip=True, separator=" ") for e in all_elements],
                key=len,
                default=""
            )
            
        return ""
    
    def _crawl_website(self, url: str) -> Dict[str, str]:
        """Execute website crawling
        
        Args:
            url: Starting URL
            
        Returns:
            Dict[str, str]: Mapping of URLs to their content
        """
        num_links = 0
        crawler_result: Dict[str, str] = {}
        
        # Initialize crawl with starting URL
        primary_domain = self._get_primary_domain(url)
        self._urls_to_crawl.append((url, 1))

        while self._urls_to_crawl:
            current_url, current_depth = self._urls_to_crawl.pop(0)

            # Skip if URL meets any exclusion criteria
            if (current_url in self._visited
                or not urlparse(current_url).netloc.endswith(primary_domain)
                or current_depth > self.max_depth
                or num_links >= self.max_links):
                continue

            self._visited.add(current_url)
            time.sleep(random.uniform(1, 3))  # Rate limiting

            try:
                # Fetch and parse page content
                response = httpx.get(current_url, timeout=20)
                response.raise_for_status()
                soup = BeautifulSoup(response.content, "html.parser")
                main_content = self._extract_main_content(soup)
                if main_content:
                    crawler_result[current_url] = main_content
                    num_links += 1

                # Find and queue additional links if not at max depth
                if current_depth < self.max_depth:
                    for link in soup.find_all("a", href=True):
                        full_url = urljoin(current_url, link["href"])
                        parsed_url = urlparse(full_url)
                        if (parsed_url.netloc.endswith(primary_domain) 
                            and not any(parsed_url.path.endswith(ext) 
                                      for ext in [".pdf", ".jpg", ".png"])
                            and full_url not in self._visited 
                            and (full_url, current_depth + 1) not in self._urls_to_crawl):
                            self._urls_to_crawl.append((full_url, current_depth + 1))

            except Exception as e:
                print(f"error:{e}")
                continue

        return crawler_result



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2025/1/6 17:21
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: au_trace_context.py

import threading
import uuid
from typing import Optional


class AuTraceContext:
    def __init__(self, session_id: Optional[str] = None,
                 trace_id: Optional[str] = None,
                 span_id: Optional[str] = None):
        self._session_id = session_id
        self._trace_id = trace_id or self._generate_id()
        self._span_id = span_id or '0'
        self._span_id_counter = 0
        self._lock = threading.Lock()

    @classmethod
    def new_context(cls):
        return cls()

    @staticmethod
    def _generate_id() -> str:
        return uuid.uuid4().hex

    @property
    def session_id(self) -> str:
        return self._session_id

    @property
    def trace_id(self) -> str:
        return self._trace_id

    @property
    def span_id(self) -> str:
        return self._span_id

    def set_session_id(self, session_id: str):
        self._session_id = session_id

    def set_trace_id(self, trace_id: str):
        self._trace_id = trace_id

    def set_span_id(self, span_id: str):
        self._span_id = span_id

    def gen_child_span_id(self) -> str:
        with self._lock:
            child_span_id = self.span_id + '.' + str(self._span_id_counter)
            self._span_id_counter += 1
            return child_span_id


    def to_dict(self) -> dict:
        return {
            "session_id": self.session_id,
            "trace_id": self.trace_id,
            "span_id": self.span_id
        }

    def __str__(self):
        return f"Context(session_id={self.session_id}, trace_id={self.trace_id}, span_id={self.span_id})"

    def __repr__(self):
        return self.__str__()



# !/usr/bin/env python3
# -*- coding:utf-8 -*-
import asyncio
from typing import Any, Optional, Literal, List

from mcp.types import CallToolResult

from agentuniverse.agent.action.tool.tool import Tool, ToolInput
from agentuniverse.base.config.component_configer.component_configer import \
    ComponentConfiger
from agentuniverse.base.config.component_configer.configers.tool_configer import \
    ToolConfiger
from agentuniverse.base.context.mcp_session_manager import MCPSessionManager, \
    MCPTempClient
from agentuniverse.base.util.async_util import run_async_from_sync


# @Time    : 2025/4/14 18:11
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: mcp_tool.py


class MCPTool(Tool):

    server_name: str = ''
    transport: Literal["stdio", "sse", "websocket", "streamable_http"] = "stdio"
    url: str = ''
    command: str = ''
    args: List[str] = []
    env: Optional[dict] = None
    connection_kwargs: Optional[dict] = None
    # You can use origin_tool_name while you want another name for this aU tool
    origin_tool_name: str = ''

    @property
    def tool_name(self) -> str:
        return self.origin_tool_name if self.origin_tool_name else self.name

    def execute(self, **kwargs) -> CallToolResult:
        session = MCPSessionManager().get_mcp_server_session_sync(
            server_name=self.server_name,
            **self.get_mcp_server_connect_args()
        )
        result = MCPSessionManager().run_async(session.call_tool, self.tool_name, kwargs)
        return result

    async def async_execute(self, **kwargs) -> CallToolResult:
        session = await MCPSessionManager().get_mcp_server_session(
            server_name=self.server_name,
            **self.get_mcp_server_connect_args()
        )
        result = await session.call_tool(self.tool_name, kwargs)
        return result


    def get_mcp_server_connect_args(self) -> dict:
        if self.transport == "sse":
            connect_args = {
                'transport': self.transport,
                'url': self.url
            }
        elif self.transport == "stdio":
            return {
                'transport': self.transport,
                "command": self.command,
                "args": self.args,
                'env': self.env
            }
        elif self.transport == "streamable_http":
            connect_args = {
                'transport': self.transport,
                'url': self.url
            }
        elif self.transport == "websocket":
            connect_args = {
                'transport': self.transport,
                'url': self.url
            }
        else:
            raise Exception(
                f'Unsupported mcp server type: {self.transport}')
        if self.connection_kwargs and isinstance(self.connection_kwargs, dict):
            connect_args.update(self.connection_kwargs)
        return connect_args

    def initialize_by_component_configer(self, component_configer: ToolConfiger) -> 'MCPTool':
        """Initialize the agent model by component configer."""
        super().initialize_by_component_configer(component_configer)
        self._initialize_by_component_configer(component_configer)
        return self

    async def get_tool_info(self):
        async with MCPTempClient(
            self.get_mcp_server_connect_args()
        ) as client:
            tools_list = await client.session.list_tools()
        tools = tools_list.tools
        tool_info = None
        for _tool in tools:
            if _tool.name == self.tool_name:
                tool_info = _tool
                break
        if not tool_info:
            raise Exception(f'No tool named {self.tool_name} in mcp server {self.server_name}')
        self.input_keys = tool_info.inputSchema['required']
        self.args_model_schema = tool_info.inputSchema
        if not self.description:
            self.description = f'{tool_info.description}\n{str(tool_info.inputSchema)}'

    def _initialize_by_component_configer(self, component_configer: ComponentConfiger) -> 'MCPTool':
        if not self.server_name:
            # use an unique name to manage session
            self.server_name = self.name
        coro = self.get_tool_info()
        run_async_from_sync(coro, timeout=10)
        return self



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/10/10 19:45
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: sql_alchemy_memory_storage.py
import datetime
import json
import uuid
from abc import abstractmethod
from typing import Optional, List, Any

from pydantic import BaseModel, ConfigDict
from sqlalchemy.orm import declarative_base, sessionmaker, Session
from sqlalchemy import Integer, String, DateTime, Text, Column, Index, and_, func, or_, create_engine, Engine, insert

from agentuniverse.agent.memory.conversation_memory.conversation_message import ConversationMessage
from agentuniverse.agent.memory.conversation_memory.enum import ConversationMessageEnum, ConversationMessageSourceType
from agentuniverse.agent.memory.memory_storage.memory_storage import MemoryStorage
from agentuniverse.agent.memory.message import Message
from agentuniverse.base.config.component_configer.component_configer import ComponentConfiger
from agentuniverse.database.sqldb_wrapper import SQLDBWrapper


class BaseMemoryConverter(BaseModel):
    """ The base class for memory converter used for converting between aU Message and SQLAlchemy model.

    Attributes:
        model_class: The SQLAlchemy model class.
        model_config: The model configuration.
    """

    model_class: Any = None
    model_config = ConfigDict(protected_namespaces=())

    @abstractmethod
    def from_sql_model(self, sql_message: Any) -> ConversationMessage:
        """Convert a SQLAlchemy model to a Message instance."""
        raise NotImplementedError

    @abstractmethod
    def to_sql_model(self, message: ConversationMessage, **kwargs) -> Any:
        """Convert a Message instance to a SQLAlchemy model."""
        raise NotImplementedError

    @abstractmethod
    def get_sql_model_class(self) -> Any:
        """Get the SQLAlchemy model class."""
        raise NotImplementedError


def create_memory_model(table_name: str, DynamicBase: Any) -> Any:
    """
    Create a memory model for a given table name.

    Args:
        table_name: The name of the table to use.
        DynamicBase: The base class to use for the model.

    Returns:
        The model class.
    """

    class MemoryModel(DynamicBase):
        """The default memory model for SqlAlchemyMemory."""
        __tablename__ = table_name
        id = Column(Integer, primary_key=True, autoincrement=True)
        session_id = Column(String(100), default='')
        content = Column(Text)
        trace_id = Column(String(100), default='')
        source = Column(String(50), default='')
        source_type = Column(String(50), default='')
        target = Column(String(50), default='')
        target_type = Column(String(50), default='')
        type = Column(String(50), default='')
        prefix = Column(String(200), default='')
        timestamp = Column(DateTime, default=func.now())
        params = Column(Text)
        pair_id = Column(String(50), default=0)
        message_id = Column(String(100), unique=True)
        additional_args = Column(Text)

        __table_args__ = (
            Index(f"idx_{table_name}_session_id_source", 'session_id', 'source', 'source_type'),
            Index(f"idx_{table_name}_session_id_source_type", 'session_id', 'target', 'target_type'),
            Index(f"idx_{table_name}_session_id_gmt_created", 'session_id', 'timestamp'),
            Index(f"idx_{table_name}_message_id_unique", 'message_id', unique=True)
        )

    return MemoryModel


class DefaultMemoryConverter(BaseMemoryConverter):
    """The default memory converter for SqlAlchemyMemory."""

    def __init__(self, table_name: str, **kwargs: Any):
        super().__init__(**kwargs)
        self.model_class = create_memory_model(table_name, declarative_base())

    def from_sql_model(self, sql_message: Any) -> Message:
        """Convert a SQLAlchemy model to a Message instance."""
        return ConversationMessage.from_dict({'id': sql_message.message_id,
                                              'conversation_id': sql_message.session_id,
                                              'source': sql_message.source,
                                              'source_type': sql_message.source_type,
                                              'target': sql_message.target,
                                              'target_type': sql_message.target_type,
                                              'content': sql_message.content,
                                              'metadata': {
                                                  'prefix': sql_message.prefix,
                                                  'timestamp': sql_message.timestamp,
                                                  'params': sql_message.params,
                                                  'pair_id': sql_message.pair_id,
                                                  'gmt_created': sql_message.timestamp
                                              },
                                              'type': sql_message.type,
                                              'trace_id': sql_message.trace_id,
                                              'additional_args': json.loads(sql_message.additional_args)
                                              })

    def to_sql_model(self, message: ConversationMessage, session_id: str = None, **kwargs) -> Any:
        """Convert a Message instance to a SQLAlchemy model."""
        return self.model_class(
            session_id=session_id, content=message.content,
            trace_id=message.trace_id,
            source=message.source,
            source_type=message.source_type,
            target=message.target,
            target_type=message.target_type,
            type=message.type,
            prefix=message.metadata.get('prefix'),
            timestamp=message.metadata.get('timestamp', datetime.datetime.now()),
            params=message.metadata.get('params'),
            pair_id=message.metadata.get('pair_id'),
            message_id=message.id or uuid.uuid4().hex,
            additional_args=json.dumps(message.additional_args)
        )

    def get_sql_model_class(self) -> Any:
        """Get the SQLAlchemy model class."""
        return self.model_class


class SqliteMemoryStorage(MemoryStorage):
    """SqlAlchemyMemoryStorage class that stores messages in a SQL database.

    Attributes:
        sqldb_table_name (str): The name of the table to store for the memory.
        sqldb_wrapper_name (str): The name of the SQLDBWrapper to use for the memory.
        memory_converter (BaseMemoryConverter): The memory converter to use for the memory.
        _sqldb_wrapper (SQLDBWrapper): The SQLDBWrapper instance to use for the memory.
    """

    sqldb_table_name: Optional[str] = 'memory'
    sqldb_path: Optional[str] = None
    memory_converter: BaseMemoryConverter = None
    engine: Optional[Engine] = None
    session: Optional[Any] = None

    model_config = {
        "arbitrary_types_allowed": True,  # 允许任意类型
    }

    def _new_client(self):
        self.engine = create_engine(self.sqldb_path, echo=False)
        self.session = sessionmaker(bind=self.engine)
        self._init_db()

    def _initialize_by_component_configer(self,
                                          memory_storage_config: ComponentConfiger) -> 'SqlAlchemyMemoryStorage':
        """Initialize the SqlAlchemyMemoryStorage by the ComponentConfiger object.

        Args:
            memory_storage_config(ComponentConfiger): A configer contains sql_alchemy_memory_storage basic info.
        Returns:
            SqlAlchemyMemoryStorage: A SqlAlchemyMemoryStorage instance.
        """
        super()._initialize_by_component_configer(memory_storage_config)
        if getattr(memory_storage_config, 'sql_table_name', None):
            self.sqldb_table_name = memory_storage_config.sql_table_name
        if getattr(memory_storage_config, 'sqldb_path', None):
            self.sqldb_path = memory_storage_config.sqldb_path
        if self.sqldb_path is None:
            raise Exception('`sqldb_wrapper_name` is not set')
        # initialize the memory converter if not set
        if self.memory_converter is None:
            self.memory_converter = DefaultMemoryConverter(self.sqldb_table_name)
        self._new_client()
        return self

    def _init_db(self) -> None:
        self._create_table_if_not_exists()

    def _create_table_if_not_exists(self) -> None:
        """Create the db table if it does not exist."""
        with self.engine.connect() as conn:
            if not conn.dialect.has_table(conn, self.sqldb_table_name):
                self.memory_converter.get_sql_model_class().__table__.create(conn)

    def delete(self, session_id: str = None, agent_id: str = None, trace_id: str = None, **kwargs) -> None:
        """Delete the memory from the database.

        Args:
            session_id (str): The session id of the memory to delete.
            agent_id (str): The agent id of the memory to delete.
        """
        if self.engine is None:
            self._init_db()
        if session_id is None and agent_id is None:
            return
        with self.session() as session:
            model_class = self.memory_converter.get_sql_model_class()
            query = session.query(model_class)
            # construct query based on the provided session_id and agent_id
            if session_id is not None:
                query = query.filter(getattr(model_class, 'session_id') == session_id)
            if agent_id is not None:
                source_col = getattr(model_class, 'source')
                type_col = getattr(model_class, 'type')
                source_type_col = getattr(model_class, 'source_type')
                source_condition = and_(source_col == agent_id,
                                        type_col == ConversationMessageEnum.OUTPUT.value,
                                        source_type_col == ConversationMessageSourceType.AGENT.value
                                        )
                target_col = getattr(model_class, 'target')
                target_type_col = getattr(model_class, 'target_type')

                target_condition = and_(target_col == agent_id,
                                        type_col == ConversationMessageEnum.INPUT.value,
                                        target_type_col == ConversationMessageSourceType.AGENT.value)
                agent_id_col = or_(source_condition, target_condition)

                query.filter(agent_id_col)
            if trace_id is not None:
                query.filter(getattr(model_class, 'trace_id') == trace_id)

            # execute delete and commit the session
            query.delete(synchronize_session=False)
            session.commit()

    def add(self, message_list: List[ConversationMessage], session_id: str = None, agent_id: str = None,
            **kwargs) -> None:
        """Add messages to the memory db.

        Args:
            message_list (List[Message]): The list of messages to add.
            session_id (str): The session id of the memory to add.
            agent_id (str): The agent id of the memory to add.
        """
        message_list = ConversationMessage.check_and_convert_message(message_list, session_id)
        if self.engine is None:
            self._init_db()
        if message_list is None:
            return

        with self.session() as session:
            for message in message_list:
                existing_message = session.query(self.memory_converter.get_sql_model_class()).filter_by(
                    message_id=message.id).first()
                if not existing_message:
                    session.add(self.memory_converter.to_sql_model(message=message,
                                                                   session_id=session_id if session_id else None,
                                                                   agent_id=agent_id, **kwargs))
            session.commit()

    def get(self, session_id: str = None, agent_id: str = None, top_k=20, trace_id: str = None, **kwargs) -> List[
        ConversationMessage]:
        """Get messages from the memory db.

        Args:
            session_id (str): The session id of the memory to get.
            agent_id (str): The agent id of the memory to get.
            top_k (int): The number of messages to get.

        Returns:
            List[Message]: The list of messages retrieved from the memory.
        """
        if self.session is None:
            self._init_db()
        with self.session() as session:
            # get the messages from the memory by session_id and agent_id
            model_class = self.memory_converter.get_sql_model_class()
            conditions = []

            # conditionally add session_id to the query
            if session_id:
                session_id_col = getattr(model_class, 'session_id')
                conditions.append(session_id_col == session_id)

            source_col = getattr(model_class, 'source')
            type_col = getattr(model_class, 'type')
            agent_type_col = getattr(model_class, 'source_type')
            target_col = getattr(model_class, 'target')
            target_agent_type_col = getattr(model_class, 'target_type')
            if 'type' in kwargs:
                memory_type_col = getattr(model_class, 'type')
                if isinstance(kwargs['type'], list):
                    conditions.append(memory_type_col.in_(kwargs['type']))
                elif isinstance(kwargs['type'], str):
                    conditions.append(type_col == kwargs['type'])
                else:
                    raise ValueError("type must be a list or str")
            if agent_id:
                agent_qa_col = and_(target_col == agent_id,
                                    target_agent_type_col == ConversationMessageSourceType.AGENT.value)
                if kwargs.get("memory_types", None) and len(kwargs["memory_types"]) > 0:
                    types_col = and_(source_col == agent_id,
                                     agent_type_col == ConversationMessageSourceType.AGENT.value,
                                     target_agent_type_col.in_(kwargs["memory_types"]))
                    conditions.append(or_(
                        types_col,
                        agent_qa_col
                    ))
                else:
                    conditions.append(agent_qa_col)
            if trace_id:
                trace_id_col = getattr(model_class, 'trace_id')
                conditions.append(trace_id_col == trace_id)
            # build the query with dynamic conditions
            query = session.query(self.memory_converter.model_class)
            if conditions:
                query = query.where(and_(*conditions))
            query = query.order_by(model_class.timestamp.asc())

            # Execute the query and fetch the results
            records = query.all()

            records = records[-top_k:]

            messages = []
            for record in records:
                messages.append(self.memory_converter.from_sql_model(record))
            return messages



import time
import traceback
from concurrent.futures import TimeoutError

from flask import Flask, Response, g, request, make_response, \
    copy_current_request_context
from loguru import logger
from werkzeug.exceptions import HTTPException
from werkzeug.local import LocalProxy

from agentuniverse.base.util.logging.general_logger import get_context_prefix
from agentuniverse.base.util.logging.log_type_enum import LogTypeEnum
from .request_task import RequestTask
from .thread_with_result import ThreadPoolExecutorWithReturnValue
from .web_util import request_param, service_run_queue, make_standard_response, \
    FlaskServerManager
from ..service_instance import ServiceInstance, ServiceNotFoundError
from ...base.context.context_coordinator import ContextCoordinator
from ...base.util.logging.logging_util import LOGGER


# Patch original flask request so it can be dumped by loguru.
class SerializableRequest:
    def __init__(self, method, path, args, form, headers):
        self.method = method
        self.path = path
        self.args = args
        self.form = form
        self.headers = headers

    def __repr__(self):
        return f"<SerializableRequest method={self.method} path={self.path}>"


def localproxy_reduce_ex(self, protocol):
    real_obj = self._get_current_object()
    return (
        SerializableRequest,
        (real_obj.method, real_obj.path, dict(real_obj.args), dict(real_obj.form), dict(real_obj.headers)),
    )


LocalProxy.__reduce_ex__ = localproxy_reduce_ex


# log stream response
def timed_generator(generator, start_time, context_prefix):
    try:
        for data in generator:
            yield data
    finally:
        elapsed_time = time.time() - start_time
        logger.bind(
            log_type=LogTypeEnum.flask_response,
            flask_response="Stream finished",
            elapsed_time=elapsed_time,
            context_prefix=context_prefix
        ).info("Stream finished.")


app = Flask(__name__)
app.config['JSON_AS_ASCII'] = False
app.json.ensure_ascii = False


@app.before_request
def before():
    logger.bind(
        log_type=LogTypeEnum.flask_request,
        flask_request=request,
        context_prefix=get_context_prefix()
    ).info("Before request.")
    g.start_time = time.time()


@app.after_request
def after_request(response):
    if not response.mimetype == "text/event-stream":
        logger.bind(
            log_type=LogTypeEnum.flask_response,
            flask_response=response,
            elapsed_time=time.time() - g.start_time,
            context_prefix=get_context_prefix()
        ).info("After request.")
    return response

@app.teardown_request
def teardown_resource(exception):
    """
    Clear the context
    """
    ContextCoordinator.end_context()


@app.route("/echo")
def echo():
    return 'Welcome to agentUniverse!!!'


@app.route("/liveness")
def liveness():
    return make_standard_response(success=True,
                                  result="liveness health check pass!")


@app.route("/service_run", methods=['POST'])
@request_param
def service_run(service_id: str, params: dict, saved: bool = False):
    """Synchronous invocation of an agent service.

    Request Args:
        service_id(`str`): The id of the agent service.
        params(`dict`): Json style params passed to service.
        saved(`bool`): Save the request and result into database.

    Return:
        Returns a dict containing two keys: success and result.
        success: This key holds a boolean value indicating the task was
            successfully or not.
        result: This key points to a nested dictionary that includes the
            result of the task.
    """
    try:
        params = {} if params is None else params
        request_task = RequestTask(ServiceInstance(service_id).run, saved,
                                   **params)
        with ThreadPoolExecutorWithReturnValue() as executor:
            future = executor.submit(copy_current_request_context(request_task.run))
            result = future.result(timeout=FlaskServerManager().sync_service_timeout)
    except TimeoutError:
        return make_standard_response(success=False,
                                      message="AU sync service timeout",
                                      status_code=504)

    return make_standard_response(success=True, result=result,
                                  request_id=request_task.request_id)


@app.route("/service_run_stream", methods=['POST'])
@request_param
def service_run_stream(service_id: str, params: dict, saved: bool = False):
    """Synchronous invocation of an agent service, return in stream form.

    Request Args:
        service_id(`str`): The id of the agent service.
        params(`dict`): Json style params passed to service.
        saved(`bool`): Save the request and result into database.

    Return:
        A SSE(Server-Sent Event) stream.
    """
    params = {} if params is None else params
    params['service_id'] = service_id
    params['streaming'] = True
    task = RequestTask(service_run_queue, saved, **params)
    context_prefix = get_context_prefix()
    response = Response(timed_generator(task.stream_run(), g.start_time, context_prefix), mimetype="text/event-stream")
    response.headers['X-Request-ID'] = task.request_id
    return response


@app.route("/service_run_async", methods=['POST'])
@request_param
def service_run_async(service_id: str, params: dict, saved: bool = True):
    """Async invocation of an agent service, return the request id used to
    get result later.

    Request Args:
        service_id(`str`): The id of the agent service.
        params(`dict`): Json style params passed to service.
        saved(`bool`): Save the request and result into database.

    Return:
        Returns a dict containing two keys: success and result.
        success: This key holds a boolean value indicating the task was
            successfully or not.
        request_id: Stand for a single request taski, can be used in
            service_run_result api to get the result of async task.
    """
    params = {} if params is None else params
    params['service_id'] = service_id
    task = RequestTask(service_run_queue, saved, **params)
    task.async_run()
    return make_standard_response(success=True,
                                  request_id=task.request_id)


@app.route("/service_run_result", methods=['GET'])
@request_param
def service_run_result(request_id: str):
    """Get the async service result.

    Request Args:
        request_id(`str`): Request id returned by async run api.

    Return:
        Returns a dict containing two keys: success and result if request_id
        exists in database.
        success: This key holds a boolean value indicating the task was
            successfully or not.
        result: This key points to a nested dictionary that includes the
            result of the task.
    """
    data = RequestTask.query_request_state(request_id)
    if data is None:
        return make_standard_response(
            success=False,
            message=f"request {request_id} not found"
        )
    return make_standard_response(success=True, result=data,
                                  request_id=request_id)


@app.errorhandler(HTTPException)
def handle_http_exception(e):
    """A global exception handler handle flask origin http exceptions."""
    response = e.get_response()
    return response


@app.errorhandler(Exception)
def handle_exception(e):
    """A global non http exception handler"""
    LOGGER.error(traceback.format_exc())
    if isinstance(e, ServiceNotFoundError):
        return make_standard_response(success=False,
                                      message=str(e),
                                      status_code=404)
    return make_standard_response(success=False,
                                  message="Internal Server Error",
                                  status_code=500)


@app.route("/chat/completions", methods=['POST'])
@request_param
def openai_protocol_chat(model: str, messages: list):
    """
    OpenAI chat completion API.
    """
    stream = request.json.get("stream", False)
    params = {
        "service_id": model,
        "messages": messages,
        "stream": stream
    }
    if stream:
        task = RequestTask(service_run_queue, False, **params)
        context_prefix = get_context_prefix()
        response = Response(timed_generator(task.user_stream_run(), g.start_time, context_prefix), mimetype="text/event-stream")
        response.headers['X-Request-ID'] = task.request_id
        return response
    try:
        params = {} if params is None else params
        request_task = RequestTask(ServiceInstance(params.get('service_id')).run, False,
                                   **params)
        with ThreadPoolExecutorWithReturnValue() as executor:
            future = executor.submit(request_task.run)
            result = future.result(timeout=FlaskServerManager().sync_service_timeout)
    except TimeoutError:
        return make_standard_response(success=False,
                                      message="AU sync service timeout",
                                      status_code=504)
    response = make_response(result, 200)
    response.headers['X-Request-ID'] = request_task.request_id
    return response



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2025/4/10 17:40
# @Author  : zhaoyifei
# @Email   : 2179709293@qq.com
# @FileName: feishu_reader.py

from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from bs4 import BeautifulSoup
import time
from typing import Dict,List
from agentuniverse.agent.action.knowledge.store.document import Document

class PublicFeishuReader:
    """Feishu public document reader using Selenium

    Extracts content from public Feishu documents through web scraping with dynamic rendering support.

    Attributes:
        driver: Selenium WebDriver instance configured for headless browsing
    """

    def __init__(self):
        """Initialize Selenium WebDriver with headless configuration"""
        chrome_options = Options()
        chrome_options.add_argument("--headless")
        chrome_options.add_argument("--disable-gpu")
        chrome_options.add_argument("--no-sandbox")
        chrome_options.add_argument(
            "User-Agent=Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36"
        )
        self.driver = webdriver.Chrome(options=chrome_options)

    def _fetch_document_content(self, url: str) -> str:
        """Fetch Feishu document content after dynamic rendering

        Args:
            url (str): URL of the public Feishu document

        Returns:
            str: Extracted document content or empty string on failure

        Raises:
            RuntimeError: If page loading fails
        """
        try:
            self.driver.get(url)
            # Wait for dynamic content loading
            time.sleep(5)
            page_source = self.driver.page_source
            soup = BeautifulSoup(page_source, 'html.parser')
            return self._parse_content(soup)
        except Exception as e:
            print(f"Error fetching document: {e}")
            return ""

    def _parse_content(self, soup: BeautifulSoup) -> str:
        """Parse document content and extract meaningful text

        Args:
            soup (BeautifulSoup): Parsed HTML document object

        Returns:
            str: Cleaned text content with deduplication

        Raises:
            ValueError: If document structure is invalid
        """
        # Find the body tag
        body = soup.find('body')
        if not body:
            return "Body tag not found"

        content = []

        # Extract document title
        title_tag = soup.find('h1')
        if title_tag:
            title_text = title_tag.get_text(strip=True)
            if title_text:
                content.append(f"Title: {title_text}")

        # Define irrelevant class names for filtering
        irrelevant_classes = ['footer', 'nav', 'navigation', 'header', 'comment', 'help', 'login']

        # Try to locate main content container
        content_div = soup.find('div', class_='doc-content')  # Update class name as needed
        if content_div:
            # Extract text from content container
            for tag in content_div.find_all(['p', 'div', 'span'], recursive=True):
                if tag.get('class') and any(cls.lower() in irrelevant_classes for cls in tag.get('class')):
                    continue
                text = tag.get_text(strip=True)
                if text and text.strip() and not any(
                        keyword in text.lower() for keyword in ['login', 'sign up', 'comment', 'help']):
                    content.append(text)
        else:
            # Fallback extraction from body
            for div in body.find_all('div', recursive=True):
                if div.get('class') and any(cls.lower() in irrelevant_classes for cls in div.get('class')):
                    continue
                text = div.get_text(strip=True)
                if text and text.strip() and not any(
                        keyword in text.lower() for keyword in ['login', 'sign up', 'comment', 'help']):
                    content.append(text)

        # Deduplicate while preserving order
        unique_content = list(dict.fromkeys(content))
        return "\n".join(unique_content) if unique_content else "No meaningful content found"

    def load_data(self, url: str) -> List[Document]:
        """Load and process Feishu document data

        Args:
            url (str): URL of the target document

        Returns:
            List[Document]: List of documents containing feishu online file content
        """
        content = self._fetch_document_content(url)
        # If the content is empty, return an empty list
        if not content:
            return []
        # Construct metadata
        metadata = {"source": url}
        # Create a Document object
        document = Document(text=content, metadata=metadata)
        # Return a list containing a single Document
        return [document]

    def __del__(self):
        """Cleanup resources by closing browser instance"""
        if hasattr(self, 'driver'):
            self.driver.quit()

"""
if __name__ == "__main__":
    # Example usage
    test_url = "https://www.feishu.cn/docx/SV2JdR5P4odO8AxIhpBcVo1Xncg"
    reader = PublicFeishuReader()
    data = reader.load_data(test_url)
    print(data)
    print("Document Summary:", data[0].text)
"""


# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2025/3/28 10:51
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: dashscope_llm_channel.py
from typing import Optional

from agentuniverse.base.config.component_configer.component_configer import ComponentConfiger
from agentuniverse.llm.llm_channel.llm_channel import LLMChannel


class DashscopeLLMChannel(LLMChannel):
    channel_api_base: Optional[str] = 'https://dashscope.aliyuncs.com/compatible-mode/v1'

    def _initialize_by_component_configer(self, component_configer: ComponentConfiger) -> 'DashscopeLLMChannel':
        super()._initialize_by_component_configer(component_configer)
        return self



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2025/1/3 14:13
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: trace_manager.py
from contextvars import ContextVar

from agentuniverse.base.annotation.singleton import singleton
from agentuniverse.base.util.tracing.au_trace_context import AuTraceContext


@singleton
class AuTraceManager:
    def __init__(self, context_class=None):
        self.context_class = context_class or AuTraceContext
        self.context_instance = ContextVar("__au_trace_context__")

    def set_context_class(self, context_class):
        self.context_class = context_class

    def recover_trace(self, trace_context):
        self.context_instance.set(trace_context)

    def reset_trace(self):
        self.context_instance.set(None)

    @property
    def trace_context(self) -> AuTraceContext:
        context = self.context_instance.get(None)
        if not context:
            context = self.context_class.new_context()
            self.context_instance.set(context)
        return context

    def get_trace_dict(self) -> dict:
        trace_dict = {}
        if self.trace_context.session_id:
            trace_dict["session_id"] = self.trace_context.session_id
        if self.trace_context.trace_id:
            trace_dict["trace_id"] = self.trace_context.trace_id
        if self.trace_context.span_id:
            trace_dict["span_id"] = self.trace_context.span_id
        return trace_dict

    def set_session_id(self, session_id):
        self.trace_context.set_session_id(session_id)

    def get_session_id(self):
        return self.trace_context.session_id

    def set_trace_id(self, trace_id):
        self.trace_context.set_trace_id(trace_id)

    def get_trace_id(self):
        return self.trace_context.trace_id

    def set_span_id(self, span_id):
        self.trace_context.set_span_id(span_id)

    def get_span_id(self):
        return self.trace_context.span_id



# !/usr/bin/env python3
# -*- coding:utf-8 -*-
import asyncio
# @Time    : 2024/3/13 14:29
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: tool.py
from typing import List, Literal, Optional

from agentuniverse.agent.action.tool.mcp_tool import MCPTool
from agentuniverse.agent.action.tool.tool_manager import ToolManager
from agentuniverse.agent.action.toolkit.toolkit import Toolkit
from agentuniverse.base.config.component_configer.component_configer import \
    ComponentConfiger
from agentuniverse.base.context.mcp_session_manager import MCPTempClient
from agentuniverse.base.util.async_util import run_async_from_sync


class MCPToolkit(Toolkit):
    transport: Literal["stdio", "sse", "websocket", "streamable_http"] = "stdio"
    url: str = ''
    command: str = ''
    args: List[str] = []
    env: Optional[dict] = None
    server_name: str = ''
    always_refresh: bool = False
    connection_kwargs: Optional[dict] = None


    def get_mcp_server_connect_args(self) -> dict:
        if self.transport == "sse":
            connect_args = {
                'transport': self.transport,
                'url': self.url
            }
        elif self.transport == "stdio":
            return {
                'transport': self.transport,
                "command": self.command,
                "args": self.args,
                'env': self.env
            }
        elif self.transport == "streamable_http":
            connect_args = {
                'transport': self.transport,
                'url': self.url
            }
        elif self.transport == "websocket":
            connect_args = {
                'transport': self.transport,
                'url': self.url
            }
        else:
            raise Exception(
                f'Unsupported mcp server type: {self.transport}')
        if self.connection_kwargs and isinstance(self.connection_kwargs, dict):
            connect_args.update(self.connection_kwargs)
        return connect_args

    def _refresh_tool_info(self):
        pass

    @property
    def tool_names(self) -> list:
        if self.always_refresh:
            self._refresh_tool_info()
        return [f'{self.name}@{tool_name}' for tool_name in self.include]


    @property
    def tool_descriptions(self) -> list:
        if self.always_refresh:
            self._refresh_tool_info()
        tools = [ToolManager().get_instance_obj(f'{self.name}@{tool_name}', new_instance=False) for tool_name in self.include]
        tool_descriptions = [f'tool name:{tool.name}\ntool description:{tool.description}\n' for tool in tools]
        return tool_descriptions

    @property
    def func_call_list(self) -> list:
        raise NotImplementedError


    async def get_all_tools(self):
        async with MCPTempClient(
            self.get_mcp_server_connect_args()
        ) as client:
            tools_list = await client.session.list_tools()
        tools = tools_list.tools
        if not self.include:
            self.include = [tool.name for tool in tools]

        for tool in tools:
            if tool.name not in self.include:
                continue
            tool_name = f'{self.name}@{tool.name}'
            tool_instance = MCPTool(
                name=tool_name,
                description=f'{tool.description}\n{str(tool.inputSchema)}',
                server_name=self.server_name,
                origin_tool_name=tool.name,
                args_model_schema=tool.inputSchema,
                input_keys=tool.inputSchema['required'],
                **self.get_mcp_server_connect_args()
            )
            ToolManager().register(tool_instance.get_instance_code(), tool_instance)

    def _initialize_by_component_configer(self, component_configer: ComponentConfiger) -> 'MCPToolkit':
        if not self.server_name:
            self.server_name = self.name
        coro = self.get_all_tools()
        run_async_from_sync(coro, 60)
        return self





# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    :
# @Author  :
# @Email   :
# @FileName: bing_search_tool.py
from typing import Optional

from pydantic import Field
from langchain_community.utilities import BingSearchAPIWrapper

from agentuniverse.agent.action.tool.common_tool.mock_search_tool import MockSearchTool
from agentuniverse.agent.action.tool.tool import Tool, ToolInput
from agentuniverse.base.util.env_util import get_from_env



class BingSearchTool(Tool):
    """The demo bing search tool.

    Implement the execute method of demo bing search tool, using the `BingSearchAPIWrapper` to implement a simple Bing search.
    """

    bing_subscription_key: Optional[str] = Field(default_factory=lambda: get_from_env("BING_SUBSCRIPTION_KEY"))
    bing_search_url: Optional[str] = Field(default='https://api.bing.microsoft.com/v7.0/search')

    def execute(self, input: str):
        if self.bing_subscription_key is None:
            return MockSearchTool().execute(input)
        query = input
        # get top5 results from Bing search.
        search = BingSearchAPIWrapper(bing_subscription_key=self.bing_subscription_key, k=5,
                                      bing_search_url=self.bing_search_url)
        return search.run(query=query)




# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/12/10 17:07
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: custom_flask_request_sink.py
from agentuniverse.base.util.logging.log_sink.flask_response_log_sink import FlaskResponseLogSink


class CustomFlaskResponseSink(FlaskResponseLogSink):
    def generate_log(self, flask_response, elapsed_time) -> str:
        if isinstance(flask_response, str):
            response_str = (f"Response: {flask_response} "
                            f"Duration: {elapsed_time:.3f}s")
        else:
            response_str = (f"Response: {flask_response.status_code} {flask_response.content_type} "
                            f"Duration: {elapsed_time:.3f}s")


            if flask_response.data:  # 记录响应体
                try:
                    response_str += f' Data:{flask_response.get_data(as_text=True)}'
                except Exception as e:
                    pass
        return response_str


# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/12/9 18:01
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: base_log_sink.py
from loguru import logger

from agentuniverse.base.config.component_configer.component_configer import \
    ComponentConfiger
from agentuniverse.base.util.logging.log_sink.log_sink import LogSink
from agentuniverse.base.util.logging.logging_config import LoggingConfig
from agentuniverse.base.util.logging.logging_util import _get_log_file_path


class BaseFileLogSink(LogSink):

    file_prefix: str = None
    log_rotation: str = LoggingConfig.log_rotation
    log_retention: str = LoggingConfig.log_retention
    compression: str = None

    def process_record(self, record):
        raise NotImplementedError("Subclasses must implement process_record.")

    def filter(self, record):
        if not record['extra'].get('log_type') == self.log_type:
            return False
        self.process_record(record)
        return True

    def register_sink(self):
        if self.sink_id == -1:
            self.sink_id = logger.add(
                sink=_get_log_file_path(self.file_prefix),
                level=self.level,
                format=self.format,
                filter=self.filter,
                rotation=self.log_rotation,
                retention=self.log_retention,
                compression=self.compression,
                encoding="utf-8",
                enqueue=self.enqueue
            )

    def _initialize_by_component_configer(self,
                                          log_sink_configer: ComponentConfiger) -> 'LogSink':
        if hasattr(log_sink_configer, "file_prefix"):
            self.file_prefix = log_sink_configer.file_prefix
        if hasattr(log_sink_configer, "log_rotation"):
            self.log_rotation = log_sink_configer.log_rotation
        if hasattr(log_sink_configer, "log_retention"):
            self.log_retention = log_sink_configer.log_retention
        if hasattr(log_sink_configer, "compression"):
            self.compression = log_sink_configer.compression
        return self



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2025/3/15 20:34
# @Author  : wangyapei
# @FileName: tavily_tool.py
# @Need install: pip install tavily-python

from typing import Any, Dict, Literal, Optional, List

from pydantic import Field

from agentuniverse.agent.action.tool.tool import Tool, ToolInput
from agentuniverse.agent.action.tool.enum import ToolTypeEnum
from agentuniverse.base.util.env_util import get_from_env
from agentuniverse.base.util.logging.logging_util import LOGGER

try:
    from tavily import TavilyClient
except ImportError:
    raise ImportError("`tavily-python` not installed. Please install using `pip install tavily-python`")


class TavilyTool(Tool):
    """Tavily搜索工具，用于通过Tavily API进行网络搜索和网页内容提取。
    
    该工具可以执行两种操作：
    1. 搜索：返回结构化的搜索结果
    2. 网页提取：从指定URL提取网页内容
    
    Args:
        api_key (Optional[str]): Tavily API密钥，如果不提供则从环境变量获取
        search_depth (Literal["basic", "advanced"]): 搜索深度，可选"basic"或"advanced"
        max_results (int): 返回的最大结果数
        include_answer (bool): 是否包含AI生成的摘要答案
        mode (Literal["search", "extract"]): 操作模式，可选"search"或"extract"
    """
    description: str = "使用Tavily API进行网络搜索和网页内容提取，获取实时在线信息"

    # 基本配置参数
    api_key: Optional[str] = Field(default_factory=lambda: get_from_env("TAVILY_API_KEY"), description="Tavily API密钥")
    search_depth: Literal["basic", "advanced"] = Field(default="basic", description="搜索深度")
    include_answer: bool = Field(default=False, description="是否包含AI生成的摘要答案")
    mode: Literal["search", "extract"] = Field(default="search", description="操作模式：search或extract")
    
    # 搜索模式可选参数
    topic: Optional[Literal["general", "news"]] = Field(default=None, description="搜索主题：general或news")
    days: Optional[int] = Field(default=None, description="当topic为news时，指定搜索的天数")
    time_range: Optional[Literal["day", "week", "month", "year"]] = Field(default=None, description="搜索时间范围")
    max_results: int = Field(default=5, description="返回的最大结果数")
    include_domains: Optional[List[str]] = Field(default=None, description="包含的域名列表")
    exclude_domains: Optional[List[str]] = Field(default=None, description="排除的域名列表")
    include_raw_content: bool = Field(default=False, description="是否包含原始内容")
    include_images: bool = Field(default=False, description="是否包含图片")
    include_image_descriptions: bool = Field(default=False, description="是否包含图片描述")
    
    # 提取模式可选参数
    extract_depth: Literal["basic", "advanced"] = Field(default="advanced", description="提取深度")

    def execute(self, input: str | list = None, **kwargs):
        """执行Tavily工具。
        
        Args:
            tool_input (ToolInput): 包含查询/URL和可选配置参数的输入对象
        
        Returns:
            Dict: 包含搜索结果或提取内容的字典
        """
        # 检查API密钥
        if not self.api_key:
            return {"error": "未提供Tavily API密钥，请设置TAVILY_API_KEY环境变量或在调用时提供api_key参数"}
            
        # 更新可选配置（如果提供）
        possible_params = [
            "api_key", "search_depth", "include_answer", "mode",
            "topic", "days", "time_range", "max_results", "include_domains", "exclude_domains",
            "include_raw_content", "include_images", "include_image_descriptions", "extract_depth"
        ]
        
        for param in possible_params:
            if kwargs.get(param) is not None:
                setattr(self, param, kwargs.get(param))

        try:
            # 初始化Tavily客户端
            client = TavilyClient(api_key=self.api_key)
            
            # 根据操作模式执行不同的操作
            if self.mode == "extract":
                # 提取模式
                urls = input
                if not urls:
                    return {"error": "未提供要提取内容的URL"}
                
                # 如果提供的是单个URL字符串，转换为列表
                if isinstance(urls, str):
                    urls = [urls]
                
                # 提取参数
                extract_params = {
                    "urls": urls,
                    "include_images": self.include_images,
                    "extract_depth": self.extract_depth
                }
                
                # 执行提取
                result = client.extract(**extract_params)
                # LOGGER.info(f"Tavily提取结果: {result}")
                
                # 直接返回API响应
                return result
            else:
                # 搜索模式
                query = input
                if not query:
                    return {"error": "未提供搜索查询"}
                
                # 搜索参数
                search_params = {
                    "query": query,
                    "search_depth": self.search_depth
                }
                
                # 添加其他可选参数（如果有设置）
                for param_name, param_attr in {
                    "topic": self.topic,
                    "days": self.days,
                    "time_range": self.time_range,
                    "max_results": self.max_results,
                    "include_domains": self.include_domains,
                    "exclude_domains": self.exclude_domains,
                    "include_answer": self.include_answer,
                    "include_raw_content": self.include_raw_content,
                    "include_images": self.include_images,
                    "include_image_descriptions": self.include_image_descriptions
                }.items():
                    if param_attr is not None:
                        search_params[param_name] = param_attr
                
                # 执行标准搜索
                result = client.search(**search_params)
                # LOGGER.info(f"Tavily搜索结果: {result}")
                
                # 直接返回API响应
                return result
                
        except Exception as e:
            LOGGER.error(f"Tavily操作出错: {e}")
            return {"error": f"操作过程中出错: {str(e)}"}



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2025/4/8 14:07
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: deepseek_official_llm_channel.py
from typing import Optional

from agentuniverse.llm.llm_channel.llm_channel import LLMChannel

DEEPSEEK_MAX_CONTEXT_LENGTH = {
    "deepseek-chat": 64000,
    "deepseek-coder": 32000,
    "deepseek-reasoner": 64000
}


class DeepseekOfficialLLMChannel(LLMChannel):
    channel_api_base: Optional[str] = 'https://api.deepseek.com/v1'

    def max_context_length(self) -> int:
        if super().max_context_length():
            return super().max_context_length()
        return DEEPSEEK_MAX_CONTEXT_LENGTH.get(self.channel_model_name, 8000)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/12 19:36
# @Author  : heji
# @Email   : lc299034@antgroup.com
# @FileName: agent.py
import json
import uuid
from abc import abstractmethod, ABC
from datetime import datetime
from threading import Thread
from typing import Optional, Any, List

from langchain_core.runnables import RunnableSerializable, RunnableConfig
from langchain_core.utils.json import parse_json_markdown

from agentuniverse.agent.action.knowledge.knowledge import Knowledge
from agentuniverse.agent.action.knowledge.knowledge_manager import \
    KnowledgeManager
from agentuniverse.agent.action.knowledge.store.document import Document
from agentuniverse.agent.action.tool.tool import Tool
from agentuniverse.agent.action.tool.tool_manager import ToolManager
from agentuniverse.agent.action.toolkit.toolkit_manager import ToolkitManager
from agentuniverse.agent.agent_model import AgentModel
from agentuniverse.agent.input_object import InputObject
from agentuniverse.agent.memory.memory import Memory
from agentuniverse.agent.memory.memory_manager import MemoryManager
from agentuniverse.agent.memory.message import Message
from agentuniverse.agent.output_object import OutputObject
from agentuniverse.agent.plan.planner.planner import Planner
from agentuniverse.agent.plan.planner.planner_manager import PlannerManager
from agentuniverse.agent.plan.planner.react_planner.stream_callback import \
    InvokeCallbackHandler
from agentuniverse.base.annotation.trace import trace_agent
from agentuniverse.base.component.component_base import ComponentBase
from agentuniverse.base.component.component_enum import ComponentEnum
from agentuniverse.base.config.application_configer.application_config_manager \
    import ApplicationConfigManager
from agentuniverse.base.config.component_configer.configers.agent_configer \
    import AgentConfiger
from agentuniverse.base.context.framework_context_manager import \
    FrameworkContextManager
from agentuniverse.base.util.agent_util import process_agent_llm_config
from agentuniverse.base.util.common_util import stream_output
from agentuniverse.base.util.logging.logging_util import LOGGER
from agentuniverse.base.util.memory_util import generate_messages, get_memory_string
from agentuniverse.base.util.system_util import process_dict_with_funcs, is_system_builtin
from agentuniverse.base.util.tracing.au_trace_manager import AuTraceManager
from agentuniverse.llm.llm import LLM
from agentuniverse.llm.llm_manager import LLMManager
from agentuniverse.prompt.chat_prompt import ChatPrompt
from agentuniverse.prompt.prompt import Prompt
from agentuniverse.prompt.prompt_manager import PromptManager
from agentuniverse.prompt.prompt_model import AgentPromptModel


class Agent(ComponentBase, ABC):
    """The parent class of all agent models, containing only attributes."""

    agent_model: Optional[AgentModel] = None

    def __init__(self):
        """Initialize the AgentModel with the given keyword arguments."""
        super().__init__(component_type=ComponentEnum.AGENT)

    @abstractmethod
    def input_keys(self) -> list[str]:
        """Return the input keys of the Agent."""
        pass

    @abstractmethod
    def output_keys(self) -> list[str]:
        """Return the output keys of the Agent."""
        pass

    @abstractmethod
    def parse_input(self, input_object: InputObject, agent_input: dict) -> dict:
        """Agent parameter parsing.

        Args:
            input_object (InputObject): input parameters passed by the user.
            agent_input (dict): agent input preparsed by the agent.
        Returns:
            dict: agent input parsed from `input_object` by the user.
        """
        pass

    @abstractmethod
    def parse_result(self, agent_result: dict) -> dict:
        """Agent result parser.

        Args:
            agent_result(dict): The raw result of the agent.
        Returns:
            dict: The parsed result of the agent
        """
        pass

    def update_trace_context(self, input_object: InputObject):
        session_id = input_object.get_data("session_id")
        if session_id:
            AuTraceManager().trace_context.set_session_id(session_id)

        trace_id = input_object.get_data("trace_id")
        if trace_id:
            AuTraceManager().trace_context.set_trace_id(trace_id)

    @trace_agent
    def run(self, **kwargs) -> OutputObject:
        """Agent instance running entry.

        Returns:
            OutputObject: Agent execution result
        """
        self.input_check(kwargs)
        input_object = InputObject(kwargs)

        self.update_trace_context(input_object)

        agent_input = self.pre_parse_input(input_object)

        planner_result = self.execute(input_object, agent_input)

        agent_result = self.parse_result(planner_result)

        self.output_check(agent_result)
        output_object = OutputObject(agent_result)
        return output_object

    @trace_agent
    async def async_run(self, **kwargs) -> OutputObject:
        self.input_check(kwargs)
        input_object = InputObject(kwargs)
        self.update_trace_context(input_object)

        agent_input = self.pre_parse_input(input_object)

        agent_result = await self.async_execute(input_object, agent_input)

        agent_result = self.parse_result(agent_result)

        self.output_check(agent_result)
        output_object = OutputObject(agent_result)
        return output_object

    def execute(self, input_object: InputObject, agent_input: dict) -> dict:
        """Execute agent instance.

        Args:
            input_object (InputObject): input parameters passed by the user.
            agent_input (dict): agent input parsed from `input_object` by the user.

        Returns:
            dict: planner result generated by the planner execution.
        """
        planner_base: Planner = PlannerManager().get_instance_obj(self.agent_model.plan.get('planner').get('name'))
        planner_result = planner_base.invoke(self.agent_model, agent_input, input_object)
        return planner_result

    async def async_execute(self, input_object: InputObject, agent_input: dict) -> dict:
        pass

    def pre_parse_input(self, input_object) -> dict:
        """Agent execution parameter pre-parsing.

        Args:
            input_object (InputObject): input parameters passed by the user.
        Returns:
            dict: agent input preparsed by the agent.
        """
        agent_input = dict()
        agent_input['chat_history'] = input_object.get_data('chat_history') or ''
        agent_input['background'] = input_object.get_data('background') or ''
        agent_input['image_urls'] = input_object.get_data('image_urls') or []
        agent_input['audio_url'] = input_object.get_data('audio_url') or ''
        agent_input['date'] = datetime.now().strftime('%Y-%m-%d')
        agent_input['session_id'] = input_object.get_data('session_id') or ''
        agent_input['agent_id'] = self.agent_model.info.get('name', '')
        self.parse_input(input_object, agent_input)
        return agent_input

    def get_instance_code(self) -> str:
        """Return the full name of the agent."""
        appname = ApplicationConfigManager().app_configer.base_info_appname
        name = self.agent_model.info.get('name')
        return (f'{appname}.'
                f'{self.component_type.value.lower()}.{name}')

    def input_check(self, kwargs: dict):
        """Agent parameter check."""
        for key in self.input_keys():
            if key not in kwargs.keys():
                raise Exception(f'Input must have key: {key}.')

    def output_check(self, kwargs: dict):
        """Agent result check."""
        if not isinstance(kwargs, dict):
            raise Exception('Output type must be dict.')
        for key in self.output_keys():
            if key not in kwargs.keys():
                raise Exception(f'Output must have key: {key}.')

    def initialize_by_component_configer(self, component_configer: AgentConfiger) -> 'Agent':
        """Initialize the Agent by the AgentConfiger object.

        Args:
            component_configer(AgentConfiger): the ComponentConfiger object
        Returns:
            Agent: the Agent object
        """
        agent_config: Optional[AgentConfiger] = component_configer.load()
        info: Optional[dict] = agent_config.info
        profile: Optional[dict] = agent_config.profile
        profile = process_dict_with_funcs(profile, component_configer.yaml_func_instance)
        profile = process_agent_llm_config(info.get('name'), profile, component_configer.default_llm_configer)
        plan: Optional[dict] = agent_config.plan
        memory: Optional[dict] = agent_config.memory
        memory = process_dict_with_funcs(memory, component_configer.yaml_func_instance)
        action: Optional[dict] = agent_config.action
        action = process_dict_with_funcs(action, component_configer.yaml_func_instance)
        agent_model: Optional[AgentModel] = AgentModel(info=info, profile=profile,
                                                       plan=plan, memory=memory, action=action)
        self.agent_model = agent_model
        return self

    def langchain_run(self, input: str, callbacks=None, **kwargs):
        """Run the agent model using LangChain."""
        try:
            parse_result = parse_json_markdown(input)
        except Exception as e:
            LOGGER.error(f"langchain run parse_json_markdown error,input(parse_result) error({str(e)})")
            return "Error , Your Action Input is not a valid JSON string"
        output_object = self.run(**parse_result, callbacks=callbacks, **kwargs)
        result_dict = {}
        for key in self.output_keys():
            result_dict[key] = output_object.get_data(key)
        return result_dict

    async def async_langchain_run(self, input: str, callbacks=None, **kwargs):
        """Run the agent model using LangChain."""
        try:
            parse_result = parse_json_markdown(input)
        except Exception as e:
            LOGGER.error(f"langchain run parse_json_markdown error,input(parse_result) error({str(e)})")
            return "Error , Your Action Input is not a valid JSON string"
        output_object = await self.async_run(**parse_result, callbacks=callbacks, **kwargs)
        result_dict = {}
        for key in self.output_keys():
            result_dict[key] = output_object.get_data(key)
        return result_dict

    def as_langchain_tool(self):
        """Convert to LangChain tool."""
        from langchain.agents.tools import Tool
        format_dict = {}
        for key in self.input_keys():
            format_dict.setdefault(key, "input val")
        format_str = json.dumps(format_dict)

        args_description = f"""
        to use this tool,your input must be a json string,must contain all keys of {self.input_keys()},
        and the value of the key must be a json string,the format of the json string is as follows:
        ```{format_str}```
        """
        return Tool(
            name=self.agent_model.info.get("name"),
            func=self.langchain_run,
            description=self.agent_model.info.get("description") + args_description
        )

    async def async_as_langchain_tool(self):
        """Convert to LangChain tool."""
        from langchain.agents.tools import Tool
        format_dict = {}
        for key in self.input_keys():
            format_dict.setdefault(key, "input val")
        format_str = json.dumps(format_dict)

        args_description = f"""
        to use this tool,your input must be a json string,must contain all keys of {self.input_keys()},
        and the value of the key must be a json string,the format of the json string is as follows:
        ```{format_str}```
        """
        return Tool(
            name=self.agent_model.info.get("name"),
            func=self.async_langchain_run,
            description=self.agent_model.info.get("description") + args_description
        )

    def process_llm(self, **kwargs) -> LLM:
        llm_name = kwargs.get('llm_name') or self.agent_model.profile.get('llm_model', {}).get('name')
        llm: LLM = LLMManager().get_instance_obj(llm_name)
        if is_system_builtin(llm):
            LOGGER.warn("The system built-in LLM configuration YAML will be removed in the next version. "
                        "Please configure your own LLM model in the application's YAML file.")
        return llm

    def process_memory(self, agent_input: dict, **kwargs) -> Memory | None:
        memory_name = kwargs.get('memory_name') or self.agent_model.memory.get('name')
        memory: Memory = MemoryManager().get_instance_obj(memory_name)
        conversation_memory_name = kwargs.get('conversation_memory') or self.agent_model.memory.get(
            'conversation_memory')
        conversation_memory: Memory = MemoryManager().get_instance_obj(
            component_instance_name=conversation_memory_name)
        if memory is None and conversation_memory is None:
            return None
        if memory is None:
            memory = conversation_memory

        chat_history: list = agent_input.get('chat_history')
        # generate a list of temporary messages from the given chat history and add them to the memory instance.
        temporary_messages: list[Message] = generate_messages(chat_history)
        if temporary_messages:
            memory.add(temporary_messages, **agent_input)

        params: dict = dict()
        params['agent_llm_name'] = kwargs.get('llm_name') or self.agent_model.profile.get('llm_model', {}).get('name')
        return memory.set_by_agent_model(**params)

    def invoke_chain(self, chain: RunnableSerializable[Any, str], agent_input: dict, input_object: InputObject,
                     **kwargs):
        if not self.judge_chain_stream(chain):
            res = chain.invoke(input=agent_input, config=self.get_run_config())
            return res
        result = []
        for token in chain.stream(input=agent_input, config=self.get_run_config()):
            stream_output(input_object.get_data('output_stream', None), {
                'type': 'token',
                'data': {
                    'chunk': token,
                    'agent_info': self.agent_model.info
                }
            })
            result.append(token)
        return self.generate_result(result)

    async def async_invoke_chain(self, chain: RunnableSerializable[Any, str], agent_input: dict,
                                 input_object: InputObject, **kwargs):
        if not self.judge_chain_stream(chain):
            res = await chain.ainvoke(input=agent_input, config=self.get_run_config())
            return res
        result = []
        async for token in chain.astream(input=agent_input, config=self.get_run_config()):
            stream_output(input_object.get_data('output_stream', None), {
                'type': 'token',
                'data': {
                    'chunk': token,
                    'agent_info': self.agent_model.info
                }
            })
            result.append(token)
        return self.generate_result(result)

    def judge_chain_stream(self,
                           chain: RunnableSerializable[Any, str]) -> bool:
        streaming = False
        for _step in chain.steps:
            if hasattr(_step, "llm"):
                return _step.kwargs.get('streaming', _step.llm.streaming)
            if hasattr(_step, "llm_channel"):
                return _step.kwargs.get('streaming', _step.llm_channel.streaming)
        return streaming

    def generate_result(self, data: list[dict | str]):
        if isinstance(data[0], str):
            return "".join(data)
        text = [val.get('text') for val in data]
        reasoning_content = [val.get('reasoning_content', "") for val in data]
        return {
            'text': "".join(text),
            'reasoning_content': "".join(reasoning_content)
        }

    @property
    def tool_names(self) -> list:
        return self._get_tool_names()

    def _get_tool_names(self) -> list:
        tool_name_list = self.agent_model.action.get('tool', [])
        if tool_name_list is None:
            tool_name_list = []
        for toolkit_name in self.agent_model.action.get('toolkit', []):
            toolkit = ToolkitManager().get_instance_obj(toolkit_name)
            tool_name_list.extend(toolkit.tool_names)
        return tool_name_list

    def invoke_tools(self, input_object: InputObject, **kwargs) -> str:
        tool_names = kwargs.get('tool_names') or self.tool_names
        if not tool_names:
            return ''

        tool_results: list = list()

        for tool_name in tool_names:
            tool: Tool = ToolManager().get_instance_obj(tool_name)
            if tool is None:
                continue
            try:
                tool_input = {key: input_object.get_data(key) for key in tool.input_keys}
                tool_results.append(str(tool.run(**tool_input)))
            except:
                LOGGER.warn(f'Tool {tool_name} call failed, maybe invalid or lack arguments')
        return "\n\n".join(tool_results)

    async def async_invoke_tools(self, input_object: InputObject, **kwargs) -> str:
        tool_names = kwargs.get('tool_names') or self.agent_model.action.get('tool', [])
        if not tool_names:
            return ''
        tool_results: list = list()
        for tool_name in tool_names:
            tool: Tool = ToolManager().get_instance_obj(tool_name)
            if tool is None:
                continue
            tool_input = {key: input_object.get_data(key) for key in tool.input_keys}
            tool_results.append(await tool.async_run(**tool_input))
        return "\n\n".join(tool_results)

    def invoke_knowledge(self, query_str: str, input_object: InputObject, **kwargs) -> str:
        knowledge_names = kwargs.get('knowledge_names') or self.agent_model.action.get('knowledge', [])
        if not knowledge_names or not query_str:
            return ''

        knowledge_results: list = list()

        for knowledge_name in knowledge_names:
            knowledge: Knowledge = KnowledgeManager().get_instance_obj(knowledge_name)
            if knowledge is None:
                continue
            knowledge_res: List[Document] = knowledge.query_knowledge(
                query_str=query_str,
                **input_object.to_dict()
            )
            knowledge_results.append(knowledge.to_llm(knowledge_res))
        return "\n\n".join(knowledge_results)

    def process_prompt(self, agent_input: dict, **kwargs) -> ChatPrompt:
        expert_framework = agent_input.pop('expert_framework', '') or ''

        profile: dict = self.agent_model.profile

        profile_instruction = profile.get('instruction')
        profile_instruction = expert_framework + profile_instruction if profile_instruction else profile_instruction

        profile_prompt_model: AgentPromptModel = AgentPromptModel(introduction=profile.get('introduction'),
                                                                  target=profile.get('target'),
                                                                  instruction=profile_instruction)

        # get the prompt by the prompt version
        prompt_version = kwargs.get('prompt_version') or self.agent_model.profile.get('prompt_version')
        version_prompt: Prompt = PromptManager().get_instance_obj(prompt_version)

        if version_prompt is None and not profile_prompt_model:
            raise Exception("Either the `prompt_version` or `introduction & target & instruction`"
                            " in agent profile configuration should be provided.")
        if version_prompt:
            version_prompt_model: AgentPromptModel = AgentPromptModel(
                introduction=getattr(version_prompt, 'introduction', ''),
                target=getattr(version_prompt, 'target', ''),
                instruction=expert_framework + getattr(version_prompt, 'instruction', ''))
            profile_prompt_model = profile_prompt_model + version_prompt_model

        chat_prompt = ChatPrompt().build_prompt(profile_prompt_model, ['introduction', 'target', 'instruction'])
        image_urls: list = agent_input.pop('image_urls', []) or []
        if image_urls:
            chat_prompt.generate_image_prompt(image_urls)

        audio_url: str = agent_input.pop('audio_url') or ''
        if audio_url:
            chat_prompt.generate_audio_prompt(audio_url)
        return chat_prompt

    def get_memory_params(self, agent_input: dict) -> dict:
        memory_info = self.agent_model.memory
        memory_types = self.agent_model.memory.get('memory_types', None)
        prune = self.agent_model.memory.get('prune', False)
        top_k = self.agent_model.memory.get('top_k', 20)
        session_id = agent_input.get('session_id')
        agent_id = self.agent_model.info.get('name')
        if not session_id:
            session_id = FrameworkContextManager().get_context('session_id')
        if "agent_id" in memory_info:
            agent_id = memory_info.get('agent_id')
        params = {
            'session_id': session_id,
            'agent_id': agent_id,
            'prune': prune,
            'top_k': top_k
        }
        if memory_types:
            params['memory_types'] = memory_types
        if agent_input.get('input'):
            params['input'] = agent_input.get('input')
        if not self.agent_model.memory.get('name') and self.agent_model.memory.get('conversation_memory'):
            params["type"] = ['input', 'output']
        return params

    def get_run_config(self, **kwargs) -> dict:
        llm_name = kwargs.get('llm_name') or self.agent_model.profile.get('llm_model', {}).get('name')
        callbacks = [InvokeCallbackHandler(
            source=self.agent_model.info.get('name'),
            llm_name=llm_name
        )]
        return RunnableConfig(callbacks=callbacks)

    def collect_current_memory(self, collect_type: str) -> bool:
        collection_types = self.agent_model.memory.get('collection_types')
        auto_trace = self.agent_model.memory.get('auto_trace', True)
        if not auto_trace:
            return False
        if collection_types and collect_type not in collection_types:
            return False
        return True

    def load_memory(self, memory, agent_input: dict):
        if memory:
            params = self.get_memory_params(agent_input)
            LOGGER.info(f"Load memory with params: {params}")
            memory_messages = memory.get(**params)
            memory_str = get_memory_string(memory_messages, agent_input.get('agent_id'))
        else:
            return "Up to Now, No Chat History"
        agent_input[memory.memory_key] = memory_str
        return memory_str

    def add_memory(self, memory: Memory, content: Any, type: str = 'Q&A', agent_input: dict[str, Any] = {}):
        if not memory:
            return
        session_id = agent_input.get('session_id')
        if not session_id:
            session_id = FrameworkContextManager().get_context('session_id')
        agent_id = self.agent_model.info.get('name')
        message = Message(id=str(uuid.uuid4().hex),
                          source=agent_id,
                          content=content if isinstance(content, str) else json.dumps(content, ensure_ascii=False),
                          type=type,
                          metadata={
                              'agent_id': agent_id,
                              'session_id': session_id,
                              'type': type,
                              'timestamp': datetime.now(),
                              'gmt_created': datetime.now().isoformat()
                          })
        memory.add([message], session_id=session_id, agent_id=agent_id)

    def summarize_memory(self, agent_input: dict[str, Any] = {}, memory: Memory = None):
        def do_summarize(params):
            content = memory.summarize_memory(**params)
            memory.add([
                Message(
                    id=str(uuid.uuid4().hex),
                    source=self.agent_model.info.get('name'),
                    content=content,
                    type='summarize'
                )
            ], session_id=params['session_id'], agent_id=params['agent_id'])

        if memory:
            params = self.get_memory_params(agent_input)
            Thread(target=do_summarize, args=(params,)).start()

    def load_summarize_memory(self, memory: Memory, agent_input: dict[str, Any] = {}) -> str:
        if memory:
            params = self.get_memory_params(agent_input)
            params['type'] = 'summarize'
            memory_messages = memory.get(**params)
            if len(memory_messages) == 0:
                return "Up to Now, No Summarize Memory"
            else:
                return memory_messages[-1].content
        return "Up to Now, No Summarize Memory"

    def get_tool_descriptions(self) -> List[str]:
        description_list = []
        if self.agent_model.action.get('tool', []):
            for tool in self.agent_model.action.get('tool', []):
                tool_ins = ToolManager().get_instance_obj(tool)
                description_list.append(f'tool name:{tool_ins.name}\ntool description:{tool_ins.description}\n')

        if self.agent_model.action.get('toolkit', []):
            for toolkit in self.agent_model.action.get('toolkit', []):
                description_list.extend(ToolkitManager().get_instance_obj(toolkit).tool_descriptions)

        return description_list

    def get_func_call_list(self) -> List[str]:
        pass

    def create_copy(self):
        copied = self.model_copy()
        if self.agent_model is not None:
            copied.agent_model = self.agent_model.model_copy(deep=True)
        return copied



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2025/4/22 10:16
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: async_util.py

import asyncio
import queue
import threading
from queue import Queue
from typing import Coroutine, TypeVar, Any, Optional

T = TypeVar('T')


def _async_runner_thread_target(coro: Coroutine[Any, Any, T], result_queue: Queue):
    """
    Internal target function for the worker thread.

    It creates a new event loop in the current thread (the worker thread),
    runs the provided coroutine within that loop, and puts the result
    or any exception encountered into the result_queue.
    """
    try:
        result = asyncio.run(coro)
        result_queue.put(result)
    except Exception as e:
        result_queue.put(e)


def run_async_from_sync(coro: Coroutine[Any, Any, T], timeout: Optional[float] = None) -> T:
    """
    Synchronously executes an asynchronous coroutine in a separate thread
    and returns its result.

    This function is designed to be called from synchronous code, especially
    when you need to call an async function without blocking an existing
    event loop (if called from an async context via a sync function).

    It starts a new thread, which creates its own event loop to run the
    coroutine. The calling thread blocks until the result is available
    or a timeout occurs.

    Args:
        coro: The coroutine object to execute (e.g., my_async_func(arg1)).
              Ensure this is the coroutine object, not the function itself.
        timeout: Optional maximum time in seconds to wait for the result.
                 If None, waits indefinitely.

    Returns:
        The result returned by the awaited coroutine.

    Raises:
        TypeError: If the provided 'coro' argument is not a coroutine.
        TimeoutError: If the timeout is reached before the coroutine completes.
        Any Exception: Any exception raised by the coroutine during its execution
                       will be re-raised in the calling thread.
    """
    if not asyncio.iscoroutine(coro):
        raise TypeError(f"Expected a coroutine, but got {type(coro).__name__}")

    # Using a queue for thread-safe communication of the result/exception
    result_queue: Queue = Queue(maxsize=1)
    worker_thread = threading.Thread(
        target=_async_runner_thread_target,
        args=(coro, result_queue),
        daemon=True
    )
    worker_thread.start()

    try:
        result = result_queue.get(timeout=timeout)
    except queue.Empty:
        raise TimeoutError(f"Operation timed out after {timeout} seconds")
    finally:
        worker_thread.join(timeout=1.0)

    if isinstance(result, Exception):
        raise result
    else:
        return result



# !/usr/bin/env python3
# -*- coding:utf-8 -*-
from dataclasses import dataclass
from typing import Any

import opentracing
from agentuniverse.base.context.framework_context_manager import \
    FrameworkContextManager
from agentuniverse.base.context.mcp_session_manager import MCPSessionManager
from agentuniverse.base.util.tracing.au_trace_manager import AuTraceManager


# @Time    : 2025/4/15 15:35
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: context_coordinator.py


@dataclass
class ContextPack:
    framework_context: dict
    trace_context: Any
    mcp_session: dict
    opentracing_span: Any


class ContextCoordinator:

    @classmethod
    def save_context(cls) -> ContextPack:
        context_pack = ContextPack(
            framework_context=FrameworkContextManager().get_all_contexts(),
            trace_context=AuTraceManager().trace_context,
            mcp_session=MCPSessionManager().save_mcp_session(),
            opentracing_span=opentracing.tracer.active_span
        )

        return context_pack

    @classmethod
    def recover_context(cls, context_pack: ContextPack):
        for var_name, var_value in context_pack.framework_context.items():
            FrameworkContextManager().set_context(var_name, var_value)
        AuTraceManager().recover_trace(context_pack.trace_context)
        MCPSessionManager().recover_mcp_session(**context_pack.mcp_session)
        if context_pack.opentracing_span:
            opentracing.tracer.scope_manager.activate(
                context_pack.opentracing_span, finish_on_close=False
            )

    @classmethod
    def end_context(cls):
        FrameworkContextManager().clear_all_contexts()
        AuTraceManager().reset_trace()
        MCPSessionManager().safe_close_stack()



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/10/24 21:12
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: common_util.py
from __future__ import annotations

import json
import re
from queue import Queue
from typing import Any, Callable, List


def stream_output(output_stream: Queue, data: dict):
    """Add data to the output stream.

    Args:
        output_stream (Queue): The output stream.
        data (dict): The data to be streamed.
    """
    if output_stream is None:
        return
    output_stream.put_nowait(data)


def _replace_new_line(match: re.Match[str]) -> str:
    value = match.group(2)
    value = re.sub(r"\n", r"\\n", value)
    value = re.sub(r"\r", r"\\r", value)
    value = re.sub(r"\t", r"\\t", value)
    value = re.sub(r'(?<!\\)"', r"\"", value)

    return match.group(1) + value + match.group(3)


def _custom_parser(multiline_string: str) -> str:
    """
    The LLM response for `action_input` may be a multiline
    string containing unescaped newlines, tabs or quotes. This function
    replaces those characters with their escaped counterparts.
    (newlines in JSON must be double-escaped: `\\n`)
    """
    if isinstance(multiline_string, (bytes, bytearray)):
        multiline_string = multiline_string.decode()

    multiline_string = re.sub(
        r'("action_input"\:\s*")(.*?)(")',
        _replace_new_line,
        multiline_string,
        flags=re.DOTALL,
    )

    return multiline_string


def parse_partial_json(s: str, *, strict: bool = False) -> Any:
    """Parse a JSON string that may be missing closing braces.

    Args:
        s: The JSON string to parse.
        strict: Whether to use strict parsing. Defaults to False.

    Returns:
        The parsed JSON object as a Python dictionary.
    """
    # Attempt to parse the string as-is.
    try:
        return json.loads(s, strict=strict)
    except json.JSONDecodeError:
        pass

    # Initialize variables.
    new_s = ""
    stack = []
    is_inside_string = False
    escaped = False

    # Process each character in the string one at a time.
    for char in s:
        if is_inside_string:
            if char == '"' and not escaped:
                is_inside_string = False
            elif char == "\n" and not escaped:
                char = "\\n"  # Replace the newline character with the escape sequence.
            elif char == "\\":
                escaped = not escaped
            else:
                escaped = False
        else:
            if char == '"':
                is_inside_string = True
                escaped = False
            elif char == "{":
                stack.append("}")
            elif char == "[":
                stack.append("]")
            elif char == "}" or char == "]":
                if stack and stack[-1] == char:
                    stack.pop()
                else:
                    # Mismatched closing character; the input is malformed.
                    return None

        # Append the processed character to the new string.
        new_s += char

    # If we're still inside a string at the end of processing,
    # we need to close the string.
    if is_inside_string:
        new_s += '"'

    # Try to parse mods of string until we succeed or run out of characters.
    while new_s:
        final_s = new_s

        # Close any remaining open structures in the reverse
        # order that they were opened.
        for closing_char in reversed(stack):
            final_s += closing_char

        # Attempt to parse the modified string as JSON.
        try:
            return json.loads(final_s, strict=strict)
        except json.JSONDecodeError:
            # If we still can't parse the string as JSON,
            # try removing the last character
            new_s = new_s[:-1]

    # If we got here, we ran out of characters to remove
    # and still couldn't parse the string as JSON, so return the parse error
    # for the original string.
    return json.loads(s, strict=strict)


def parse_json_markdown(
        json_string: str, *, parser: Callable[[str], Any] = parse_partial_json
) -> dict:
    """
    Parse a JSON string from a Markdown string.

    Args:
        json_string: The Markdown string.
        parser: The parser to use. Defaults to `parse_partial_json`.

    Returns:
        The parsed JSON object as a Python dictionary.
    """
    try:
        return _parse_json(json_string, parser=parser)
    except json.JSONDecodeError:
        # Try to find JSON string within triple backticks
        match = re.search(r"```(json)?(.*)", json_string, re.DOTALL)

        # If no match found, assume the entire string is a JSON string
        if match is None:
            json_str = json_string
        else:
            # If match found, use the content within the backticks
            json_str = match.group(2)
    return _parse_json(json_str, parser=parser)


def _parse_json(
        json_str: str, *, parser: Callable[[str], Any] = parse_partial_json
) -> dict:
    # Strip whitespace and newlines from the start and end
    json_str = json_str.strip().strip("`")

    # handle newlines and other special characters inside the returned value
    json_str = _custom_parser(json_str)

    # Parse the JSON string into a Python dictionary
    return parser(json_str)


def parse_and_check_json_markdown(text: str, expected_keys: List[str]) -> dict:
    """
    Parse a JSON string from a Markdown string and check that it
    contains the expected keys.

    Args:
        text: The Markdown string.
        expected_keys: The expected keys in the JSON string.

    Returns:
        The parsed JSON object as a Python dictionary.
    """
    try:
        json_obj = parse_json_markdown(text)
    except json.JSONDecodeError as e:
        raise Exception(f"Got invalid JSON object. Error: {e}")
    for key in expected_keys:
        if key not in json_obj:
            raise Exception(
                f"Got invalid return object. Expected key `{key}` "
                f"to be present, but got {json_obj}"
            )
    return json_obj




# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/12/5 17:43
# @Author  : weizjajj 
# @Email   : weizhongjie.wzj@antgroup.com
# @FileName: conversation_message.py

import uuid
from typing import Optional, List

from langchain_core.messages import AIMessage, SystemMessage, HumanMessage
from pydantic import Field

from agentuniverse.agent.memory.enum import ChatMessageEnum
from langchain_core.prompts.chat import BaseStringMessagePromptTemplate

from agentuniverse.agent.memory.conversation_memory.enum import ConversationMessageSourceType, ConversationMessageEnum
from agentuniverse.agent.memory.message import Message
from agentuniverse.base.context.framework_context_manager import FrameworkContextManager


class ConversationMessage(Message):
    """
    The basic class for conversation memory message

    Attributes:
        id (Optional[str]): Unique identifier.
        trace_id (Optional[str]): Trace ID.
        conversation_id (Optional[str]): Conversation ID.
        source (Optional[str]): Message source.
        source_type (Optional[str]): Type of the message source.
        target (Optional[str]): Message target.
        target_type (Optional[str]): Type of the message target.
        type (Optional[str]): Message type.
        content (Optional[str]): Message content.
        metadata (Optional[dict]): The metadata of the message.
    """
    id: Optional[str | int] = uuid.uuid4().hex
    trace_id: Optional[str] = None
    conversation_id: Optional[str] = None
    source: Optional[str] = None
    source_type: Optional[str] = None
    target: Optional[str] = None
    target_type: Optional[str] = None
    type: Optional[str] = None
    content: Optional[str] = None
    metadata: Optional[dict] = None
    additional_args: Optional[dict] = Field(default_factory=dict)

    @staticmethod
    def as_langchain_list(message_list: List['ConversationMessage']):
        """Convert agentUniverse(aU) message list to langchain message list """
        messages = []
        for message in message_list:
            if message.type == ChatMessageEnum.SYSTEM.value or message.type == ChatMessageEnum.HUMAN.value or message.type == ChatMessageEnum.AI.value:
                messages.append(message)
                continue
            # only got agent message
            if message.target_type != ConversationMessageSourceType.AGENT.value:
                continue
            if message.source_type not in [ConversationMessageSourceType.AGENT.value,
                                           ConversationMessageSourceType.USER.value]:
                continue
            if message.source_type == ConversationMessageSourceType.AGENT.value and message.type == ConversationMessageEnum.OUTPUT.value:
                messages.append(message)
            elif message.target_type == ConversationMessageSourceType.AGENT.value and message.type == ConversationMessageEnum.INPUT.value:
                messages.append(message)
        return [message.as_langchain() for message in messages]

    def as_langchain(self):
        """Convert the agentUniverse(aU) message class to the langchain message class."""
        if self.type == ConversationMessageEnum.INPUT.value:
            return HumanMessage(content=self.content)
        elif self.type == ConversationMessageEnum.OUTPUT.value:
            return AIMessage(content=self.content)
        elif self.type == ChatMessageEnum.SYSTEM.value:
            return SystemMessage(content=self.content)
        elif self.type == ChatMessageEnum.HUMAN.value:
            return HumanMessage(content=self.content)
        elif self.type == ChatMessageEnum.AI.value:
            return AIMessage(content=self.content)
        else:
            return BaseStringMessagePromptTemplate.from_template(self.content)

    @classmethod
    def from_dict(cls, data: dict):
        """Convert the agentUniverse(aU) message class to the dict."""
        return cls(**data)

    @classmethod
    def from_message(cls, message: Message, session_id: str):
        if not message.metadata:
            message.metadata = {}
        message.metadata['prefix'] = '之前对话的摘要：' if message.type == 'summarize' else ''
        message.metadata['params'] = "{}"
        trace_id = message.metadata.get('trace_id')
        if not trace_id:
            trace_id = FrameworkContextManager().get_context('trace_id')
            message.metadata['trace_id'] = trace_id
        return cls(
            id=uuid.uuid4().hex,
            content=message.content,
            metadata=message.metadata,
            type=message.type,
            source=message.source,
            source_type='agent',
            target=message.source,
            target_type='agent',
            trace_id=trace_id,
            conversation_id=message.metadata.get('session_id') if not session_id else session_id,
        )

    @classmethod
    def check_and_convert_message(cls, messages, session_id: str = None):
        if len(messages) == 0:
            return []
        message = messages[0]
        if isinstance(message, cls):
            return messages
        if isinstance(message, Message):
            return [cls.from_message(m, session_id) for m in messages]



# !/usr/bin/env python3
# -*- coding:utf-8 -*-
import functools
import inspect
from typing import Literal

from agentuniverse.agent.action.tool.tool import Tool
from agentuniverse.agent.action.tool.tool_manager import ToolManager
from agentuniverse.agent.action.toolkit.toolkit_manager import ToolkitManager
from agentuniverse.base.annotation.singleton import singleton
from agentuniverse.base.component.component_enum import ComponentEnum
from agentuniverse.base.config.component_configer.component_configer import \
    ComponentConfiger

# @Time    : 2025/5/8 17:37
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: mcp_server_manager.py


DEFAULT_SERVER_NAME = 'default_mcp_server'


def is_method_overridden(
        override_method: callable,
        origin_method: callable) -> bool:
    if not callable(override_method):
        return False

    is_overridden = False
    if origin_method is None:
        is_overridden = True
    elif override_method is not origin_method:
        is_overridden = True

    return is_overridden


def create_exposed_tool_method_wrapper(tool_name):
    method_name = 'async_execute'
    original_method = getattr(ToolManager().get_instance_obj(tool_name).__class__, method_name)

    # If async_execute method hasn't been overridden, use execute method
    if not is_method_overridden(
        original_method,
        getattr(Tool, method_name)
    ):
        method_name = 'execute'
        original_method = getattr(ToolManager().get_instance_obj(tool_name).__class__, method_name)

    original_signature = inspect.signature(original_method)
    original_parameters = list(original_signature.parameters.values())

    exposed_parameters = original_parameters[1:]

    exposed_signature = inspect.Signature(
        parameters=exposed_parameters,
        return_annotation=original_signature.return_annotation
    )

    if inspect.iscoroutinefunction(original_method):
        @functools.wraps(original_method)
        async def wrapper(*args, **kwargs):
            instance = ToolManager().get_instance_obj(tool_name)
            return await getattr(instance, method_name)(*args, **kwargs)
    else:
        @functools.wraps(original_method)
        def wrapper(*args, **kwargs):
            instance = ToolManager().get_instance_obj(tool_name)
            return getattr(instance, method_name)(*args, **kwargs)
    wrapper.__signature__ = exposed_signature

    return wrapper


@singleton
class MCPServerManager:
    server_tool_map: dict = {
        DEFAULT_SERVER_NAME: {
            'tool': [],
            'toolkit': []
        }
    }

    def register_mcp_tool(self, configer_instance: ComponentConfiger, configer_type: str):
        if isinstance(configer_instance.as_mcp_tool, dict):
            mcp_tool_config = configer_instance.as_mcp_tool
            server_name = mcp_tool_config.get('server_name', DEFAULT_SERVER_NAME)
            if server_name not in self.server_tool_map:
                self.server_tool_map[server_name] = {
                    'tool': [],
                    'toolkit': []
                }
        elif isinstance(configer_instance.as_mcp_tool, bool) and configer_instance.as_mcp_tool:
            server_name = DEFAULT_SERVER_NAME
        else:
            return

        if configer_type == ComponentEnum.TOOL.value:
            self.server_tool_map[server_name]['tool'].append(configer_instance.name)
        else:
            self.server_tool_map[server_name]['toolkit'].append(configer_instance.name)

    def start_server(self,
                     host: str = '0.0.0.0',
                     port: int = 8890,
                     transport: Literal["sse", "streamable_http"] = "sse"):
        import contextlib
        import uvicorn
        from fastapi import FastAPI
        from mcp.server.fastmcp import FastMCP
        from starlette.applications import Starlette
        from starlette.routing import Mount
        mcp_server_list = []

        for server_name, tool_dict in self.server_tool_map.items():
            mcp_server = FastMCP(server_name)

            for _tool in set(tool_dict['tool']):
                tool_instance = ToolManager().get_instance_obj(_tool)
                mcp_server.add_tool(
                    fn=create_exposed_tool_method_wrapper(_tool),
                    name=tool_instance.name,
                    description=tool_instance.description
                )

            for _toolkit in set(tool_dict['toolkit']):
                toolkit_instance = ToolkitManager().get_instance_obj(_toolkit)
                for _tool in toolkit_instance.tool_names:
                    tool_instance = ToolManager().get_instance_obj(_tool)
                    mcp_server.add_tool(
                        fn=create_exposed_tool_method_wrapper(_tool),
                        name=tool_instance.name,
                        description=tool_instance.description
                    )
            mcp_server_dict = {
                'server_name': server_name,
                'server': mcp_server
            }
            if transport == 'sse':
                mcp_server_dict['mount'] = Mount(f"/{server_name}", app=mcp_server.sse_app())

            mcp_server_list.append(mcp_server_dict)

        if transport == 'sse':
            app = Starlette(routes=[_mcp_server['mount'] for _mcp_server in mcp_server_list])
        elif transport == 'streamable_http':
            @contextlib.asynccontextmanager
            async def lifespan(app: FastAPI):
                async with contextlib.AsyncExitStack() as stack:
                    for _mcp_server in mcp_server_list:
                        await stack.enter_async_context(_mcp_server['server'].session_manager.run())
                    yield

            app = FastAPI(lifespan=lifespan)
            for _mcp_server in mcp_server_list:
                app.mount(f"/{_mcp_server['server_name']}", _mcp_server['server'].streamable_http_app())
        else:
            raise Exception('Unsupported mcp server type')

        uvicorn.run(app, host=host, port=port)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2025/2/16 17:00
# @Author  : zhouxiaoji
# @Email   : zh_xiaoji@qq.com
# @FileName: arxiv_tool.py
from typing import Optional, Any, List
import os
from dataclasses import dataclass
from enum import Enum
from pydantic import Field
from agentuniverse.agent.action.tool.tool import Tool, ToolInput
from agentuniverse.base.annotation.retry import retry


class SearchMode(Enum):
    SEARCH = "search"   
    DETAIL = "detail"  


@dataclass
class PaperSummary:
    paper_id: str
    title: str
    authors: List[str]
    publish_date: str
    summary: str
    pdf_url: str


class ArxivTool(Tool):
    
    sch_engine: Optional[Any] = None
    MAX_QUERY_LENGTH: int = Field(default=300, description="查询字符串最大长度")

    def execute(self, input: str, mode: str):
        try:
            import arxiv
        except ImportError:
            raise ImportError("arxiv is required. Install with: pip install arxiv")

        if self.sch_engine is None:
            self.sch_engine = arxiv.Client()

        if mode not in [m.value for m in SearchMode]:
            raise ValueError(f"Invalid mode: {mode}. Must be one of {[m.value for m in SearchMode]}")

        query = input
        return (self.find_papers_by_str(query) if mode == SearchMode.SEARCH.value
                else self.retrieve_full_paper_text(query))
        
    def _process_query(self, query: str) -> str:
        if len(query) <= self.MAX_QUERY_LENGTH:
            return query
        
        words: List[str] = query.split()
        processed_words: List[str] = []
        current_length: int = 0
        for word in words:
            word_length = len(word) + 1 
            if current_length + word_length <= self.MAX_QUERY_LENGTH:
                processed_words.append(word)
                current_length += word_length
            else:
                break
        return ' '.join(processed_words)

    @retry(3, 1.0)
    def find_papers_by_str(self, query) -> str:
        processed_query = self._process_query(query)
        result_num:int = 10   
        try:
            import arxiv
        except ImportError:
            raise ImportError("arxiv is required. Install with: pip install arxiv")
    
        search = arxiv.Search(
            query="abs:" + processed_query,
            max_results=result_num,
            sort_by=arxiv.SortCriterion.Relevance)

        papers: List[PaperSummary] = []
        for result in self.sch_engine.results(search):
            paper = PaperSummary(
                paper_id=result.pdf_url.split("/")[-1],
                title=result.title,
                authors=[str(author) for author in result.authors],
                publish_date=str(result.published).split()[0],
                summary=result.summary.replace('\n', ' '),
                pdf_url=result.pdf_url
            )
            papers.append(paper)
        return self._format_paper_results(papers)

    @retry(3, 1.0)
    def retrieve_full_paper_text(self, paper_id: str) -> str:
        try:
            import arxiv
        except ImportError:
            raise ImportError("arxiv is required. Install with: pip install arxiv")
        search = arxiv.Search(id_list=[paper_id])
        paper = next(self.sch_engine.results(search))
        paper.download_pdf(filename="downloaded-paper.pdf") 
        
        try:
            import pypdf
        except ImportError:
            raise ImportError(
                "pypdf is required to read PDF files: `pip install pypdf`"
            )
        reader = pypdf.PdfReader('downloaded-paper.pdf')
        text_content = [page.extract_text() for page in reader.pages]
        if os.path.exists("downloaded-paper.pdf"):
            os.remove("downloaded-paper.pdf")
        return "\n\n".join(text_content)

    def _format_paper_results(self, papers: List[PaperSummary]) -> str:
        if not papers:
            return "No papers found."

        formatted_results = []
        for i, paper in enumerate(papers, 1):
            paper_info = [
                f"[{i}] {paper.title}",
                f"Authors: {', '.join(paper.authors)}",
                f"Published: {paper.publish_date}",
                f"Paper ID: {paper.paper_id}",
                f"PDF URL: {paper.pdf_url}",
                f"Summary: {paper.summary}",
                "-" * 80
            ]
            formatted_results.append("\n".join(paper_info))
        return "\n\n".join(formatted_results)


# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/7/23 14:00
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: doc_processor.py

from abc import abstractmethod
from typing import List, Optional

from agentuniverse.agent.action.knowledge.store.query import Query
from agentuniverse.agent.action.knowledge.store.document import Document
from agentuniverse.base.component.component_base import ComponentEnum
from agentuniverse.base.component.component_base import ComponentBase
from agentuniverse.base.config.component_configer.component_configer import \
    ComponentConfiger


class DocProcessor(ComponentBase):
    """The basic class for doc processor.
    
    DocProcessor is an abstract base class that defines the interface for document 
    processing components in the agentUniverse framework. Document processors can 
    transform, filter, or enhance documents before they are used by agents.
    
    Attributes:
        component_type: Enum value identifying this as a document processor component.
        name: Optional name identifier for the processor.
        description: Optional description of the processor's functionality.
    """

    component_type: ComponentEnum = ComponentEnum.DOC_PROCESSOR
    name: Optional[str] = None
    description: Optional[str] = None

    class Config:
        arbitrary_types_allowed = True

    def process_docs(self, origin_docs: List[Document], query: Query = None) -> \
            List[Document]:
        """Process input documents，return should also be a document list."""
        return self._process_docs(origin_docs, query)

    @abstractmethod
    def _process_docs(self, origin_docs: List[Document],
                      query: Query = None) -> \
            List[Document]:
        """Process input documents，return should also be a document list.
        
        This is the core implementation method that subclasses must override to 
        provide specific document processing logic. The method should transform 
        the input documents according to the processor's purpose and optionally 
        use the query for context-aware processing.
        
        Args:
            origin_docs: List of documents to be processed.
            query: Optional query object that may influence the processing.
            
        Returns:
            List[Document]: Processed documents.
        """
        pass

    def _initialize_by_component_configer(self,
                                         doc_processor_configer: ComponentConfiger) \
            -> 'DocProcessor':
        """Initialize the DocProcessor by the ComponentConfiger object.

        Args:
            doc_processor_configer(ComponentConfiger): A configer contains DocProcessor
            basic info.
        Returns:
            DocProcessor: A DocProcessor instance.
        """
        self.name = doc_processor_configer.name
        self.description = doc_processor_configer.description
        return self



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2025/5/2 17:40
# @Author  : KiteSoar
# @Email   : hushihao2020x@163.com
# @FileName: yuque_reader.py

import re
import time
import random
import json
import urllib.parse
import requests
from pydantic import ConfigDict
from requests.adapters import HTTPAdapter, Retry
from bs4 import BeautifulSoup
from typing import List, Any, Optional

from agentuniverse.agent.action.knowledge.reader.reader import Reader
from agentuniverse.agent.action.knowledge.store.document import Document


class YuqueReader(Reader):
    """
    YuqueReader is a specialized reader designed to fetch and parse content from Yuque

    Attributes:
        cookies: Cookies for authentication with Yuque.
        session: HTTP session with retry support.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    cookies: Optional[str] = None
    session: Optional[requests.Session] = None
    def __init__(self, cookies: str = None, **data: Any):
        """Initialize HTTP session with retry support and optional cookies"""
        super().__init__(**data)
        self.cookies = cookies
        self.session = requests.Session()
        retries = Retry(total=5, backoff_factor=0.5, status_forcelist=[500,502,503,504])
        self.session.mount('https://', HTTPAdapter(max_retries=retries))

    def _fetch_url_title(self, url: str) -> str:
        """Fetch page title and clean illegal filename characters"""
        headers = {'Cookie': self.cookies} if self.cookies else {}
        try:
            resp = self.session.get(url, headers=headers, timeout=10)
            resp.raise_for_status()
            soup = BeautifulSoup(resp.text, 'html.parser')
            tag = soup.title
            if not tag or not tag.string:
                return "Untitled"
            title = tag.string.strip()
            title = re.sub(r'[\\/:*?"<>|]', '-', title)
            return title.replace(' · 语雀', '')
        except requests.exceptions.RequestException:
            return "Request Error"

    def _fetch_page_markdown(self, book_id: str, slug: str) -> str:
        """Fetch markdown source for a single document"""
        headers = {'Cookie': self.cookies} if self.cookies else {}
        url = f'https://www.yuque.com/api/docs/{slug}?book_id={book_id}&merge_dynamic_data=false&mode=markdown'
        resp = self.session.get(url, headers=headers, timeout=20)
        if resp.status_code != 200:
            print("The document download failed. The page may have been deleted.", book_id, slug)
            return ''
        data = resp.json().get('data', {})
        md = data.get('sourcecode', '')
        # Process image references inline
        def repl(m):
            src = m.group(1)
            return f'![]({src})'
        return re.sub(r'!\[.*?\]\((.*?)\)', repl, md)

    def _load_data(self, url: str) -> List[Document]:
        """Fetch all docs in a Yuque book and return as List[Document]"""
        headers = {'Cookie': self.cookies} if self.cookies else {}
        try:
            resp = self.session.get(url, headers=headers, timeout=10)
            resp.raise_for_status()
            encoded = re.findall(r'decodeURIComponent\("(.+)"\)\);', resp.text)[0]
            docs = json.loads(urllib.parse.unquote(encoded))
        except Exception as e:
            print(f"Request failed: {e}")
            return []

        book_title = self._fetch_url_title(url)
        # sanitize titles for metadata keys if needed
        chars = '/:*?"<>|\n\r'
        trans = str.maketrans({c: '_' for c in chars})

        documents: List[Document] = []
        for item in docs['book']['toc']:
            if item['title'] != book_title:
                continue
            md = self._fetch_page_markdown(str(docs['book']['id']), item['url'])
            if not md:
                continue
            metadata = {
                'source': url,
                'doc_title': item['title'],
                'sanitized_title': item['title'].translate(trans)
            }
            documents.append(Document(text=md, metadata=metadata))
            # Respectful delay
            time.sleep(random.uniform(1, 3))
        return documents

    def __del__(self):
        """Close HTTP session"""
        try:
            self.session.close()
        except:
            pass

# if __name__ == '__main__':
#     # Example usage
#     reader = YuqueReader(cookies=None)
#     test_url = "Fill in the publicly accessible Yuque document link"
#     docs = reader.load_data(test_url)
#     for doc in docs:
#         print(f"Title: {doc.metadata['doc_title']}")
#         print(doc.text)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2025/3/4 15:06
# @Author  : hiro
# @Email   : hiromesh@qq.com
# @FileName: metrics_types.py

from typing import TypedDict


class CodeMetrics(TypedDict):
    line_count: int
    code_line_count: int
    avg_line_length: float
    max_line_length: int
    character_count: int



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/12/10 17:07
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: custom_flask_request_sink.py
from agentuniverse.base.util.logging.log_sink.flask_request_log_sink import \
    FlaskRequestLogSink


class CustomFlaskRequestSink(FlaskRequestLogSink):
    def generate_log(self, flask_request) -> str:
        log_string = (f"Request: {flask_request.method} {flask_request.path} "
                      f"Headers: {dict(flask_request.headers)}")
        if flask_request.data:
            try:
                log_string += f" Body: {flask_request.get_data(as_text=True)}"
            except Exception as e:
                pass

        return log_string



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2025/2/14 14:47
# @Author  : weizjajj 
# @Email   : weizhongjie.wzj@antgroup.com
# @FileName: StrOutputParse.py

from typing import List

from langchain_core.output_parsers import StrOutputParser
from langchain_core.output_parsers.base import T
from langchain_core.outputs import Generation


class ReasoningOutputParser(StrOutputParser):
    def parse_result(self, result: List[Generation], *, partial: bool = False) -> T:
        if not result:
            return ""

        reasoning_text = ""
        if result[0].message.additional_kwargs:
            additional_kwargs = getattr(result[0].message, "additional_kwargs")
            if additional_kwargs and "reasoning_content" in additional_kwargs:
                reasoning_text = result[0].message.additional_kwargs.get("reasoning_content")
            return {
                "text": result[0].text,
                "reasoning_content": reasoning_text
            }
        return {
            "text": result[0].text,
        }



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/11/21 16:55
# @Author  : weizjajj 
# @Email   : weizhongjie.wzj@antgroup.com
# @FileName: trace_memory.py

import datetime
import json
import queue
import traceback
import uuid
from concurrent.futures.thread import ThreadPoolExecutor
from threading import Thread
from typing import List, Optional

from agentuniverse.agent.agent_manager import AgentManager

from agentuniverse.agent.action.knowledge.store.document import Document
from agentuniverse.agent.memory.conversation_memory.conversation_message import ConversationMessage
from agentuniverse.agent.memory.conversation_memory.enum import ConversationMessageSourceType
from agentuniverse.agent.memory.memory_manager import MemoryManager
from agentuniverse.agent.output_object import OutputObject
from agentuniverse.base.annotation.singleton import singleton
from agentuniverse.base.config.application_configer.application_config_manager import ApplicationConfigManager
from agentuniverse.base.context.framework_context_manager import FrameworkContextManager
from agentuniverse.base.util.logging.logging_util import LOGGER


def generate_relation_str(source: str, target: str, source_type: str, target_type: str, type: str):
    if source_type == 'agent' and target_type == 'agent' and type == 'input':
        return f"智能体 {source} 向智能体 {target} 提出了一个问题"
    if source_type == 'agent' and target_type == 'agent' and type == 'output':
        return f"智能体 {target} 回答了智能体 {source} 的问题"
    if source_type == 'agent' and target_type == 'tool' and type == 'input':
        return f"智能体 {source} 调用了工具 {target}，执行的参数是"
    if source_type == 'agent' and target_type == 'tool' and type == 'output':
        return f"工具 {target} 返回给智能体 {source} 的执行结果"
    if source_type == 'agent' and target_type == 'knowledge' and type == 'input':
        return f"智能体 {source} 在知识库 {target} 中进行了搜索，关键词是"
    if source_type == 'agent' and target_type == 'knowledge' and type == 'output':
        return f"知识库 {target} 返回给智能体 {source} 的搜索结果"
    if source_type == 'agent' and target_type == 'llm' and type == 'input':
        return f"智能体 {source} 向大模型 {target} 提问"
    if source_type == 'agent' and target_type == 'llm' and type == 'output':
        return f"大模型 {target} 返回给智能体 {source} 的答案"
    if source_type == 'unknown' and target_type == 'agent' and type == 'input':
        return f"未知类型 {source} 向智能体 {target} 提出了一个问题"
    if source_type == 'unknown' and target_type == 'agent' and type == 'output':
        return f"智能体 {target} 回答了未知 {source} 的问题"
    if source_type == "user" and target_type == 'agent' and type == 'input':
        return f"用户向智能体 {target} 提出了一个问题"
    if source_type == 'user' and target_type == 'agent' and type == 'output':
        return f"智能体 {target} 回答了用户的问题"
    elif type == 'input':
        return f"{source} 向 {target} 询问了一个问题"
    elif type == 'output':
        return f"{source} 回答了 {target} 的问题"
    elif type == 'summary':
        return f"{source} 的摘要"
    return None


def generate_relation_str_en(source: str, target: str, source_type: str, target_type: str, type: str):
    if source_type == 'agent' and target_type == 'agent' and type == 'input':
        return f"Agent {source} asked a question to agent {target}"
    if source_type == 'agent' and target_type == 'agent' and type == 'output':
        return f"Agent {target} answered the question asked by agent {source}"
    if source_type == 'agent' and target_type == 'tool' and type == 'input':
        return f"Agent {source} called tool {target}, the parameters are"
    if source_type == 'agent' and target_type == 'tool' and type == 'output':
        return f"Tool {target} returned the result to agent {source}"
    if source_type == 'agent' and target_type == 'knowledge' and type == 'input':
        return f"Agent {source} searched in knowledge {target}, the keywords are"
    if source_type == 'agent' and target_type == 'knowledge' and type == 'output':
        return f"Knowledge {target} returned the result to agent {target}"
    if source_type == 'agent' and target_type == 'llm' and type == 'input':
        return f"Agent {source} asked a question to llm {target}"
    if source_type == 'agent' and target_type == 'llm' and type == 'output':
        return f"LLM {target} returned the answer to agent {source}"
    if source_type == 'unknown' and target_type == 'agent' and type == 'input':
        return f"Unknown type {source} asked a question to agent {target}"
    if source_type == 'unknown' and target_type == 'agent' and type == 'output':
        return f"Agent {target} answered the unknown {source} question"
    if source_type == "user" and target_type == 'agent' and type == 'input':
        return f"User asked a question to agent {target}"
    if source_type == 'user' and target_type == 'agent' and type == 'output':
        return f"Agent {target} answered the user's question"
    if type == 'input':
        return f"{source} asked a question to {target}"
    elif type == 'output':
        return f"{target} answered {source}'s question"
    elif type == 'summary':
        return f"{source} summary"
    return None


def sync_to_sub_agent_memory(message: ConversationMessage, session_id: str, memory_name: str):
    def add_message(agent_name: str, memory_names: list, collect_type: str):
        agent_instance = AgentManager().get_instance_obj(agent_name)
        agent_memory = agent_instance.agent_model.memory.get('conversation_memory')
        collection_types = agent_instance.agent_model.memory.get('collection_types')
        if collection_types and collect_type not in collection_types:
            return
        if agent_memory:
            memory_instance = MemoryManager().get_instance_obj(agent_memory)
            memory_instance.add([message], session_id=session_id)
            memory_names.append(agent_memory)

    memory_names = [memory_name]
    if message.source_type == ConversationMessageSourceType.AGENT.value:
        add_message(message.source, memory_names, message.target_type)

    if message.target_type == ConversationMessageSourceType.AGENT.value:
        add_message(message.target, memory_names, message.source_type)


@singleton
class ConversationMemoryModule:

    def __init__(self):
        conversation_memory_configer = ApplicationConfigManager().app_configer.conversation_memory_configer
        self.instance_name = conversation_memory_configer.get('instance_name', '')
        self.activate = conversation_memory_configer.get('activate', False)
        self.logging = conversation_memory_configer.get('logging', False)
        self.collection_types = conversation_memory_configer.get('collection_types', ['agent', 'user'])
        self.conversation_format = conversation_memory_configer.get('conversation_format', 'cn')
        self.max_content_length = conversation_memory_configer.get('max_content_length', 8000)
        self.queue = queue.Queue(1000)
        self.thread_pool = ThreadPoolExecutor(max_workers=conversation_memory_configer.get('thread_pool', 4))
        Thread(target=self._consume_queue, daemon=True).start()

    def _consume_queue(self):
        while True:
            func = self.queue.get()
            try:
                self.thread_pool.submit(func)
            except Exception as e:
                LOGGER.error(f"Failed to process trace info: {e}")
                # 打印详细堆栈信息
                traceback.print_exc()
            finally:
                self.queue.task_done()

    def _add_trace_info(self, source: str,
                        source_type: str,
                        target: str,
                        target_type: str,
                        type: str,
                        params: dict, **kwargs) -> None:
        if not self.activate:
            return
        content = None
        if type == "input" and target_type == 'agent':
            agent_instance = AgentManager().get_instance_obj(target)
            input_field = agent_instance.agent_model.memory.get('input_field')
            if input_field and input_field in params:
                content = params.get(input_field)
        elif type == "output" and source_type == 'agent':
            agent_instance = AgentManager().get_instance_obj(source)
            output_field = agent_instance.agent_model.memory.get('output_field')
            if output_field and output_field in params:
                content = params.get(output_field)

        if content is None and type in params:
            content = params[type]
        elif content is None:
            try:
                content = json.dumps(params, ensure_ascii=False)
            except Exception as e:
                content = str(e)
        try:
            params_json = json.dumps(params, ensure_ascii=False)
        except Exception as e:
            params_json = json.dumps({
                "error": str(e)
            }, ensure_ascii=False)
        if self.conversation_format == 'cn':
            prefix = generate_relation_str(source, target, source_type, target_type, type)
        else:
            prefix = generate_relation_str_en(source, target, source_type, target_type, type)
        if not prefix:
            return
        if isinstance(content, str) and len(content) > self.max_content_length:
            content = content[:self.max_content_length]
        if len(params_json) > self.max_content_length:
            params_json = params_json[:self.max_content_length]
        if self.logging:
            LOGGER.info(
                f"{kwargs.get('session_id')} | {kwargs.get('trace_id')}| {kwargs.get('pair_id')} |\n {prefix}:{content}")
        message = ConversationMessage(
            id=uuid.uuid4().hex,
            conversation_id=kwargs.get('session_id'),
            trace_id=kwargs.get('trace_id'),
            source=source,
            source_type=source_type,
            target=target,
            target_type=target_type,
            type=type,
            metadata={
                "timestamp": datetime.datetime.now(),
                "prefix": prefix,
                "params": params_json,
                "pair_id": kwargs.get('pair_id'),
                'gmt_created': datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
            },
            content=f"{content}"
        )
        if self.instance_name:
            memory = MemoryManager().get_instance_obj(self.instance_name)
            if memory:
                memory.add([message], session_id=kwargs.get('session_id'))
        sync_to_sub_agent_memory(message, kwargs.get('session_id'), self.instance_name)

    def _add_trace(self, start_info, target_info: dict, type: str, params: dict, session_id: str, trace_id: str,
                   pair_id: str):
        if "kwargs" in params:
            params = params['kwargs']
        if params is str:
            params = {
                type: params
            }

        kwargs = {'source': start_info['source'], 'source_type': start_info['type'], 'target': target_info['source'],
                  'target_type': target_info['type'], 'type': type, 'params': params, 'trace_id': trace_id,
                  'session_id': session_id,
                  "pair_id": pair_id}
        self._add_trace_info(**kwargs)

    def add_trace_info(self, start_info: dict, target_info: dict, type: str, params: dict, pair_id: str):
        """Add trace info to the memory."""
        trace_id = FrameworkContextManager().get_context('trace_id')
        if trace_id is None:
            trace_id = str(uuid.uuid4())
            FrameworkContextManager().set_context('trace_id', trace_id)

        session_id = FrameworkContextManager().get_context('session_id')
        if session_id is None:
            session_id = str(uuid.uuid4())
            FrameworkContextManager().set_context('session_id', session_id)
        def add_trace():
            self._add_trace(start_info, target_info, type, params, session_id,
                            trace_id, pair_id)

        self.queue.put_nowait(add_trace)

    def add_tool_input_info(self, start_info: dict, target: str, params: dict, pair_id: str, auto: bool = True):
        """Add trace info to the memory."""

        if not self.collection_current_agent_memory(start_info, 'tool', auto):
            return

        target_info = {'source': target, 'type': 'tool'}
        self.add_trace_info(start_info, target_info, 'input', params, pair_id)

    def add_tool_output_info(self, start_info: dict, target: str, params: dict, pair_id: str, auto: bool = True):
        """Add trace info to the memory."""
        if not self.collection_current_agent_memory(start_info, 'tool', auto):
            return

        target_info = {'source': target, 'type': 'tool'}
        self.add_trace_info(start_info, target_info, 'output', params, pair_id)

    def add_knowledge_input_info(self, start_info: dict, target: str, params: dict, pair_id: str, auto: bool = True):

        if not self.collection_current_agent_memory(start_info, 'knowledge', auto):
            return

        target_info = {'source': target, 'type': 'knowledge'}
        self.add_trace_info(start_info, target_info, 'input', params, pair_id)

    def add_knowledge_output_info(self, start_info: dict, target: str, params: List[Document], pair_id: str,
                                  auto: bool = True):

        if not self.collection_current_agent_memory(start_info, 'knowledge', auto):
            return
        target_info = {'source': target, 'type': 'knowledge'}
        doc_data = []
        for doc in params:
            doc_data.append(doc.text)
        self.add_trace_info(start_info, target_info, 'output', {
            'output': "\n==============================\n".join(doc_data)
        }, pair_id)

    def add_agent_input_info(self, start_info: dict, instance: 'Agent', params: dict, pair_id: str,
                             auto: bool = True):
        if auto:
            if not instance.collect_current_memory(start_info.get('type')):
                return
            if not self.activate:
                return
            if 'agent' not in self.collection_types:
                return

        target_info = {'source': instance.agent_model.info.get('name'), 'type': 'agent'}
        input_keys = instance.input_keys()
        if "kwargs" in params:
            params: dict = params['kwargs']
            params = params.copy()
            params.pop('output_stream') if 'output_stream' in params else params
        if auto:
            params = {key: params[key] for key in input_keys}
        self.add_trace_info(start_info, target_info, 'input', params, pair_id)

    def add_agent_result_info(self, agent_instance: 'Agent', agent_result: Optional[OutputObject | dict],
                              target_info: dict,
                              pair_id: str, auto: bool = True):

        if auto:
            if not agent_instance.collect_current_memory(target_info.get('type')):
                return
            if not self.activate:
                return
            if 'agent' not in self.collection_types:
                return

        trace_id = FrameworkContextManager().get_context('trace_id')
        session_id = FrameworkContextManager().get_context('session_id')

        def add_trace():
            output_keys = agent_instance.output_keys()
            if auto:
                params = {key: agent_result.get_data(key) for key in output_keys}
            else:
                params = agent_result
            start_info = {
                "source": agent_instance.agent_model.info.get('name'),
                "type": "agent"
            }
            self._add_trace(target_info, start_info, 'output', params, session_id, trace_id, pair_id)
        self.queue.put_nowait(add_trace)

    def add_llm_input_info(self, start_info: dict, target: str, prompt: str, pair_id: str, auto=True):
        if not self.collection_current_agent_memory(start_info, 'llm', auto):
            return

        target_info = {'source': target, 'type': 'llm'}
        self.add_trace_info(start_info, target_info, 'input', {'input': prompt}, pair_id)

    def add_llm_output_info(self, start_info: dict, target: str, output: str, pair_id: str, auto=True):
        if not self.collection_current_agent_memory(start_info, 'llm', auto):
            return
        target_info = {'source': target, 'type': 'llm'}
        self.add_trace_info(start_info, target_info, 'output', {
            'output': output
        }, pair_id)

    def collection_current_agent_memory(self, info: dict, collection_type: str, auto: bool):
        if not auto:
            return True
        if not self.activate:
            return False
        if info.get('type') == 'agent':
            agent_id = info.get('source')
            agent_instance = AgentManager().get_instance_obj(agent_id)
            if agent_instance:
                collection_types = agent_instance.agent_model.memory.get('collection_types')
                res = agent_instance.collect_current_memory(collection_type)
                if not res:
                    return False
                if collection_types:
                    return res
        if collection_type not in self.collection_types:
            return False
        return True
